/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer;

import static androidx.exifinterface.media.ExifInterface.ORIENTATION_ROTATE_180;
import static com.realwear.internal.utils.Utils.isInRange;

import android.annotation.SuppressLint;
import android.app.Fragment;
import android.app.FragmentTransaction;
import android.app.PendingIntent;
import android.content.ContentResolver;
import android.content.Intent;
import android.database.Cursor;
import android.graphics.Bitmap;
import android.graphics.RectF;
import android.graphics.drawable.Drawable;
import android.media.MediaMetadataRetriever;
import android.media.MediaScannerConnection;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.CancellationSignal;
import android.os.Environment;
import android.os.Handler;
import android.os.Looper;
import android.os.RemoteException;
import android.provider.DocumentsContract;
import android.util.Log;
import android.util.SparseArray;
import android.view.View;
import android.webkit.MimeTypeMap;
import android.widget.ProgressBar;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.exifinterface.media.ExifInterface;

import com.itextpdf.text.exceptions.BadPasswordException;
import com.itextpdf.text.pdf.PdfReader;
import com.realwear.documentviewer.fragment.DocumentLoadingFragment;
import com.realwear.documentviewer.fragment.DocumentViewerFragment;
import com.realwear.documentviewer.fragment.IDocumentController;
import com.realwear.documentviewer.fragment.IFragmentController;
import com.realwear.documentviewer.fragment.ImageViewerFragment;
import com.realwear.documentviewer.fragment.SearchFragment;
import com.realwear.documentviewer.fragment.ThumbNailViewerFragment;
import com.realwear.documentviewer.fragment.ViewerFragment;
import com.realwear.documentviewer.memory.BitmapRequestHandler;
import com.realwear.documentviewer.memory.PDFPageRequestHandler;
import com.realwear.documentviewer.memory.PDFSearchHandler;
import com.realwear.documentviewer.object.Bookmark;
import com.realwear.documentviewer.utils.DataManager;
import com.realwear.documentviewer.utils.PermissionUtils;
import com.realwear.documentviewer.views.BookmarkOverlay;
import com.realwear.internal.utils.ObservableReadValue;
import com.realwear.internal.utils.ObservableWriteView;
import com.realwear.internal.utils.android.ServiceConnector;
import com.realwear.tagservice.ITagService;
import com.realwear.view.AlertDialog;
import com.realwear.view.styles.StyledActivity;
import com.squareup.picasso.Picasso;
import com.squareup.picasso.RequestCreator;
import com.squareup.picasso.Target;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import io.reactivex.rxjava3.android.schedulers.AndroidSchedulers;
import io.reactivex.rxjava3.core.Observable;
import io.reactivex.rxjava3.core.ObservableOnSubscribe;
import io.reactivex.rxjava3.core.Observer;
import io.reactivex.rxjava3.disposables.Disposable;
import io.reactivex.rxjava3.schedulers.Schedulers;

/**
 * Main activity for the application. Displays documents in fragments.
 */
@SuppressLint("LogNotTimber")
public class MainActivity extends StyledActivity implements IDocumentController, IFragmentController {
    public static final int MIN_ZOOM = 1;
    public static final int MAX_ZOOM = 5;
    static final String IS_PHOTO_EXTRA = "is_photo";
    static final String SHOW_DELETE_EXTRA = "show_delete";
    static final String SHOW_PREVIOUS_NEXT = "show_previous_next";
    static final String SORT_BY_DATE = "sort_by_date";
    private static final String EXTRA_FILE_PATH = "file_path";

    static final String JSON_PAGE = "page";
    static final String JSON_ZOOM = "zoom";

    private static final float ROTATION_CORRECTION = 180f;

    private static final String TAG = MainActivity.class.getSimpleName();

    private static final String[] IMAGE_EXTENSIONS = {".jpg", ".jpeg", ".png", ".bmp", ".webp", ".gif"};
    private static final String DOCUMENT_EXTENSION = ".pdf";

    private static final Comparator<File> BY_DATE_SORTER = Comparator.comparingLong(File::lastModified);
    private static final Comparator<File> BY_NAME_SORTER = (l, r) -> l.getPath().compareToIgnoreCase(r.getPath());

    static final String TAGS_CONFIG_FILE_LOCATION = Environment.getExternalStorageDirectory() +
            "/realwear/com.realwear.tags/config.json";

    private ThumbNailViewerFragment mThumbnailFragment;
    private DocumentViewerFragment mDocumentViewFragment;
    private ImageViewerFragment mImageViewerFragment;
    private SearchFragment mSearchFragment;
    private DocumentLoadingFragment mDocumentLoadingFragment;
    private BookmarkOverlay mBookmarkOverlay;
    private ViewerFragment mCurrentRenderer;
    private Fragment mCurrentFragment;
    private List<ViewerFragment> mViewerFragments;
    private boolean mIsBookMarkMode = false;
    private boolean mShowDelete, mShowPrevNext;
    private boolean mIsPhoto;
    private boolean mSortByDate;
    private int mGotoX, mGotoY;
    private int mGotoPage;
    private ProgressBar mSpinner;

    private Uri mFileLocation;

    private String docLocation;
    private String docMod;

    private int pageCount;
    private final ObservableWriteView<Integer> mCurrentPage = new ObservableWriteView<>(0);
    public final ObservableReadValue<Integer> currentPage = mCurrentPage.getObservableReadValue();

    private Picasso mImageLoader;

    private DataManager mDatabase;

    private boolean mCanStoreBookmarks;
    private final ServiceConnector<ITagService> mTagServiceConnector = new ServiceConnector<>(
            ITagService.Stub::asInterface, this, "com.realwear.TagService", "com.realwear.tagservice",
            this::handleTaggingServiceBeingConnected, this::handleTaggingServiceDisconnected
    );

    private final SparseArray<List<RectF>> mResultsByPage = new SparseArray<>();

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_document);

        DocumentApplication.setContext(this);

        mDatabase = new DataManager(this);

        final PDFPageRequestHandler pdfPageRequest = new PDFPageRequestHandler(getContentResolver());
        final BitmapRequestHandler mBitmapRequest = new BitmapRequestHandler(getContentResolver());

        mImageLoader = new Picasso.Builder(getApplicationContext())
                .addRequestHandler(pdfPageRequest)
                .addRequestHandler(mBitmapRequest)
                .build();

        createFragments();

        mBookmarkOverlay = findViewById(R.id.bookmark_overlay);
        mSpinner = findViewById(R.id.spinner);

        mViewerFragments = Arrays.asList(mImageViewerFragment, mDocumentViewFragment);
        for (ViewerFragment frag : mViewerFragments) {
            frag.setFragmentController(this);
        }

        if (!PermissionUtils.INSTANCE.checkAppPermissions(this)) {
            PermissionUtils.INSTANCE.requestAppPermissions(this);
            return;
        }
        onNewIntent(getIntent());
    }

    /**
     * Create all the fragments used in the document viewer.
     */
    private void createFragments() {
        mThumbnailFragment = ThumbNailViewerFragment.newInstance();
        mDocumentViewFragment = DocumentViewerFragment.newInstance();
        mImageViewerFragment = ImageViewerFragment.newInstance();
        mDocumentLoadingFragment = DocumentLoadingFragment.newInstance();

        mSearchFragment = SearchFragment.newInstance();
        mSearchFragment.setController(this);
    }

    @Override
    protected void onResume() {
        super.onResume();
        mBookmarkOverlay.onResume();
        mTagServiceConnector.connect();
    }

    @Override
    protected void onPause() {
        mTagServiceConnector.disconnect();
        mBookmarkOverlay.onPause();
        super.onPause();
    }

    @Override
    public void onBackPressed() {
        if (mBookmarkOverlay.isVisible()) {
            mBookmarkOverlay.hide();
            mDocumentViewFragment.refreshCommands();
        } else if (mCurrentFragment instanceof SearchFragment) {
            showDocumentFragment();
        } else {
            finish();
        }
    }

    @Override
    protected void onNewIntent(Intent intent) {
        super.onNewIntent(intent);

        mIsBookMarkMode = intent.getBooleanExtra("is_bookmark", false);
        mIsPhoto = intent.getBooleanExtra(IS_PHOTO_EXTRA, false);
        mShowPrevNext = intent.getBooleanExtra(SHOW_PREVIOUS_NEXT, false);
        mShowDelete = intent.getBooleanExtra(SHOW_DELETE_EXTRA, false);
        mSortByDate = intent.getBooleanExtra(SORT_BY_DATE, false);
        mGotoPage = parseIntent(intent, JSON_PAGE, 1);
        int gotoZoom = parseIntent(intent, JSON_ZOOM, 1);

        mGotoX = parseIntent(intent, "x", 0);
        mGotoY = parseIntent(intent, "y", 0);

        if (mGotoPage < 1) {
            mGotoPage = 1;
        }
        gotoZoom = clamp(gotoZoom, MIN_ZOOM, MAX_ZOOM);

        mFileLocation = intent.getData();
        final String docTruePath = intent.getStringExtra(EXTRA_FILE_PATH);
        if (docTruePath != null && !docTruePath.isEmpty()) {
            mFileLocation = Uri.fromFile(new File(docTruePath));
            mCanStoreBookmarks = true;
            handleDirectFileURI();
        } else if (mFileLocation.toString().startsWith("content://")) {
            mCanStoreBookmarks = false;
            handleContentProviderURI();
        } else {
            Toast.makeText(this, getString(R.string.unsupported_file), Toast.LENGTH_SHORT).show();
            finish();
        }
        setViewportLock(false);

        for (ViewerFragment frag : mViewerFragments) {
            frag.setFile(mFileLocation, getFilename(mFileLocation));
        }
        mDocumentViewFragment.clearHighlightRectangles();
        for (ViewerFragment viewer : mViewerFragments) {
            viewer.setZoomLevel(gotoZoom);
        }
    }

    private void handleDirectFileURI() {
        if (mFileLocation != null) {
            docLocation = mFileLocation.getEncodedPath();
            try {
                docLocation = java.net.URLDecoder.decode(docLocation, "UTF-8");
                if (docLocation.contains(":")) {
                    docLocation = docLocation.split(":", 2)[1];
                }
            } catch (UnsupportedEncodingException e) {
                Log.w(TAG, e);
            }
        } else {
            Log.e(TAG, "No file location given");
            finish();
        }

        setLocation();
    }

    /**
     * Handle a file that has been passed via a content provider.
     */
    private void handleContentProviderURI() {
        docLocation = mFileLocation.toString();
        setLocationFromContentProvider(mFileLocation);
    }

    /**
     * Get the filename from a URI.
     *
     * @param uri The URI to get the filename for.
     * @return The file name for the URI, or null if it cannot be found.
     */
    private String getFilename(Uri uri) {
        if (uri == null) {
            return null;
        }

        if (uri.getScheme() != null && uri.getScheme().equals("content")) {
            final String filename = getContentResolverData(
                    uri, DocumentsContract.Document.COLUMN_DISPLAY_NAME);
            if (filename != null) {
                return filename;
            }
        } else {
            final List<String> segments = uri.getPathSegments();
            return segments.get(segments.size() - 1);
        }

        return null;
    }

    /**
     * Extract some data from a content provided URI.
     *
     * @param uri  The content provided URI to extract the data from.
     * @param data The data to extract.
     * @return The data, or null if it cannot be found.
     */
    private String getContentResolverData(Uri uri, String data) {
        final CancellationSignal cancellationSignal = new CancellationSignal();

        final ContentResolver contentResolver = getContentResolver();
        final Cursor cursor = contentResolver.query(
                uri, new String[]{data}, new Bundle(), cancellationSignal);
        if (cursor == null) {
            return null;
        }

        try {
            while (cursor.moveToNext()) {
                final int mimeTypeIndex = cursor.getColumnIndex(data);
                return cursor.getString(mimeTypeIndex);
            }
        } catch (IllegalStateException e) {
            return null;
        }

        return null;
    }

    /**
     * Load the local current document location into the relevant renderer.
     */
    public void setLocation() {
        showDocumentLoadingFragment(mFileLocation, () -> {
            if (isImageFilename(docLocation)) {
                setupImageFragment();
            } else if (docLocation.toLowerCase().endsWith(".pdf")) {
                File file = new File(docLocation);
                docMod = String.valueOf(file.lastModified());
                setLocationFromDocument(mFileLocation);
            } else {
                Toast.makeText(
                        this, getString(R.string.unsupported_file), Toast.LENGTH_SHORT).show();
                finish();
            }
        });
    }

    /**
     * Load a content provided file into the relevant renderer.
     *
     * @param uri The URI to the content provided file.
     */
    private void setLocationFromContentProvider(Uri uri) {
        final ContentResolver contentResolver = getContentResolver();
        final MimeTypeMap mime = MimeTypeMap.getSingleton();
        final String type = "." + mime.getExtensionFromMimeType(contentResolver.getType(uri));

        showDocumentLoadingFragment(uri, () -> {
            if (isImageFilename(type)) {
                setupImageFragment();
            } else if (type.equals(DOCUMENT_EXTENSION)) {
                setLocationFromDocument(uri);
            } else {
                Toast.makeText(
                        this, getString(R.string.unsupported_file), Toast.LENGTH_SHORT).show();
                finish();
            }
        });
    }

    /**
     * Load a document into the renderer.
     *
     * @param uri The URI to the content provided document.
     */
    private void setLocationFromDocument(Uri uri) {
        Observable
                .create((ObservableOnSubscribe<Integer>) emitter -> {
                    final ContentResolver contentResolver = getContentResolver();

                    try (InputStream inputStream = contentResolver.openInputStream(uri)) {
                        if (inputStream == null) {
                            emitter.onError(new FileUnavailableException(
                                    getString(R.string.file_unavailable)));
                            return;
                        }

                        final PdfReader pdfReader = new PdfReader(inputStream);
                        emitter.onNext(pdfReader.getNumberOfPages());
                        pdfReader.close();

                        emitter.onComplete();
                    } catch (IOException e) {
                        emitter.onError(e);
                    }
                })
                .subscribeOn(Schedulers.newThread())
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe(new Observer<Integer>() {
                    @Override
                    public void onSubscribe(
                            @io.reactivex.rxjava3.annotations.NonNull Disposable d) {
                        // Nothing to be done.
                    }

                    @Override
                    public void onNext(@io.reactivex.rxjava3.annotations.NonNull Integer pages) {
                        pageCount = pages;
                    }

                    @Override
                    public void onError(@io.reactivex.rxjava3.annotations.NonNull Throwable e) {
                        if (e instanceof BadPasswordException) {
                            displayErrorToUser(R.string.open_error_passwd);
                        } else if (e instanceof FileUnavailableException ||
                                e instanceof FileNotFoundException) {
                            displayErrorToUser(R.string.file_unavailable);
                        } else {
                            displayErrorToUser(R.string.unsupported_file);
                        }

                        Log.e(TAG, "Failed to open content", e);
                        finish();
                    }

                    @Override
                    public void onComplete() {
                        if (pageCount > 0) {
                            mGotoPage = clamp(mGotoPage, 1, pageCount);
                            gotoPage(mGotoPage - 1);
                        } else {
                            gotoPage(0);
                        }
                    }
                });
    }

    /**
     * Display an error message to the user.
     *
     * @param errorStringResourceId The string resource containing the error message.
     */
    private void displayErrorToUser(int errorStringResourceId) {
        Toast.makeText(this, getString(errorStringResourceId), Toast.LENGTH_SHORT).show();
    }

    /**
     * Setup and show the image fragment on the activity.
     */
    private void setupImageFragment() {
        pageCount = 1;

        RequestCreator loadImageCreator = mImageLoader.load(BitmapRequestHandler.SCHEME_BITMAP + ":" + docLocation)
                .resize(2048, 0)
                .onlyScaleDown();

        if (Build.VERSION.SDK_INT > Build.VERSION_CODES.R) {
            imageOrientationCheck(loadImageCreator);
        }

        loadImageCreator.into(mImageTarget);
    }

    /**
     * Applies Exif rotation value to the image loader.
     * <p>
     * Before Android 12 we didn't need to apply an image rotation
     * to Picasso because, of how images were captured in the left
     * eye orientation, (bugfix)CLOUD-2537.
     *
     * @param loadImageCreator, image loader request for Picasso.
     */
    private void imageOrientationCheck(RequestCreator loadImageCreator) {
        ExifInterface exifInterface;

        try {
            exifInterface = new ExifInterface(docLocation);
        } catch (IOException ioException) {
            Log.d(TAG, "Rotation Exif data not found", ioException);
            return;
        }

        int rotationValue = exifInterface.getAttributeInt(ExifInterface.TAG_ORIENTATION, -1);
        if (rotationValue == ORIENTATION_ROTATE_180) {
            Log.d(TAG, "Applying Image rotation correction: ORIENTATION_ROTATE_180");
            loadImageCreator.rotate(ROTATION_CORRECTION);
        }
    }

    /**
     * Show the preview fragment.
     */
    public void showPreviewFragment() {
        mThumbnailFragment.setPageCount(pageCount, docLocation);
        showFragment(mThumbnailFragment, false);
    }

    /**
     * Add a fragment to the stack.
     *
     * @param fragmentTransaction The FragmentTransaction instance to use for adding the fragment.
     *                            Note that it is the caller's responsibility to call commit.
     * @param fragment            The fragment to add.
     */
    private void addFragment(FragmentTransaction fragmentTransaction, Fragment fragment) {
        if (!fragment.isAdded()) {
            fragmentTransaction.add(R.id.container, fragment);
        }
    }

    /**
     * Update the fragment that is currently displayed to the user.
     *
     * @param fragment     The new fragment to display.
     * @param hidePrevious true to call hide the previous fragment, false to leave the state alone.
     */
    private void showFragment(Fragment fragment, boolean hidePrevious) {
        if (mCurrentFragment == fragment) {
            return;
        }

        final FragmentTransaction fragmentTransaction = getFragmentManager().beginTransaction();

        if (hidePrevious && mCurrentFragment != null) {
            fragmentTransaction.remove(mCurrentFragment);
        }

        addFragment(fragmentTransaction, fragment);
        fragmentTransaction.show(fragment);
        mCurrentFragment = fragment;

        fragmentTransaction.commit();
    }

    @Override
    public void showBookmarkFragment() {
        List<Bookmark> bookmarks = mDatabase.getAllBookmarks(docLocation, docMod);
        mBookmarkOverlay.show(bookmarks, pageCount);
    }

    /**
     * Display the document fragment to the user.
     */
    public void showDocumentFragment() {
        showFragment(mDocumentViewFragment, true);
        hideBookmarks();

        mCurrentRenderer = mDocumentViewFragment;
    }

    public void showDocumentPreview(View v) {
        if (pageCount != 0) {
            if (mDocumentViewFragment.isVisible()) {
                showPreviewFragment();
            } else {
                showDocumentFragment();
            }
        }
    }

    /**
     * Show the document loading fragment.
     *
     * @param uri      The URI of the document that is being loaded.
     * @param listener A listener that will be called when the loading fragment has been displayed.
     */
    public void showDocumentLoadingFragment(Uri uri,
                                            ViewerFragment.OnFragmentLoadedListener listener) {
        String filename = getFilename(uri);
        mDocumentLoadingFragment.setFragmentLoadedListener(() -> {
            mDocumentLoadingFragment.setFilename(filename);

            if (listener != null) {
                listener.onFragmentLoaded();
            }
        });
        showFragment(mDocumentLoadingFragment, true);
    }

    private void hideBookmarks() {
        mBookmarkOverlay.hide();
    }

    public void gotoNextPage() {
        if (pageCount >= mCurrentPage.get()) {
            gotoPage(mCurrentPage.get() + 1);
        }
    }

    public void gotoPrevPage() {
        if (mCurrentPage.get() != 0) {
            gotoPage(mCurrentPage.get() - 1);
        }
    }

    private boolean areBookmarksShown() {
        return mBookmarkOverlay.isVisible();
    }

    public void refreshBookmarks() {
        if (areBookmarksShown()) {
            List<Bookmark> bookmarks = mDatabase.getAllBookmarks(docLocation, docMod);
            mBookmarkOverlay.onRefresh(bookmarks, pageCount);
        }
    }

    /**
     * Open a page in a document.
     *
     * @param page The page to open.
     */
    public void gotoPage(int page) {
        if (areBookmarksShown()) {
            hideBookmarks();
        }

        page = clamp(page, 0, pageCount - 1);

        mCurrentPage.update(page);

        mImageLoader
                .load(PDFPageRequestHandler.SCHEME_PDF_PAGE + ":" + docLocation + ":" + page)
                .resize(0, 2048)
                .onlyScaleDown()
                .into(mDocumentTarget);
    }

    /**
     * Target used for handling loading of images from documents.
     */
    protected final Target mDocumentTarget = new Target() {
        @Override
        public void onBitmapLoaded(Bitmap bitmap, Picasso.LoadedFrom from) {
            if (mCurrentFragment == mDocumentViewFragment) {
                displayBitmap(bitmap);
            } else {
                mDocumentViewFragment.setFragmentLoadedListener(() -> displayBitmap(bitmap));
                showDocumentFragment();
            }
        }

        @Override
        public void onBitmapFailed(Exception e, Drawable errorDrawable) {
            Log.d(TAG, "Bitmap failed to load", e);
            displayErrorToUser(R.string.file_unavailable);
            finish();
        }

        @Override
        public void onPrepareLoad(Drawable placeHolderDrawable) {
            // No need to do anything.
        }
    };

    /**
     * Target used for handling loading of images.
     */
    protected final Target mImageTarget = new Target() {
        @Override
        public void onBitmapLoaded(Bitmap bitmap, Picasso.LoadedFrom from) {
            mImageViewerFragment.setFragmentLoadedListener(() -> displayBitmap(bitmap));
            showFragment(mImageViewerFragment, true);

            if (mBookmarkOverlay.isVisible()) {
                hideBookmarks();
            }

            mCurrentRenderer = mImageViewerFragment;
        }

        @Override
        public void onBitmapFailed(Exception e, Drawable errorDrawable) {
            displayErrorToUser(R.string.file_unavailable);
            finish();
        }

        @Override
        public void onPrepareLoad(Drawable placeHolderDrawable) {
            // No need to do anything.
        }
    };

    /**
     * Display a bitmap in the current renderer.
     *
     * @param bitmap The bitmap to display.
     */
    private void displayBitmap(Bitmap bitmap) {
        if (mCurrentRenderer != null) {
            mCurrentRenderer.setBitmap(bitmap);
        }

        resetDocument(null);

        mDocumentViewFragment.clearHighlightRectangles();
        mDocumentViewFragment.addHighlightRectangles(mResultsByPage.get(mCurrentPage.get(),
                Collections.emptyList()));

        if (mIsBookMarkMode) {
            mCurrentRenderer.setViewportLock(true);

            Handler handler = new Handler();
            handler.postDelayed(() -> {
                if (mGotoX != 0 && mGotoY != 0) {
                    mCurrentRenderer.gotoLocation(mGotoX, mGotoY);
                }
            }, 50);

            mIsBookMarkMode = false;
        }

        mSpinner.setVisibility(View.INVISIBLE);
    }

    public void resetDocument(View v) {
        if (mCurrentRenderer != null) {
            mCurrentRenderer.resetDocument(true);
            setViewportLock(false);
        }
    }

    public static boolean isImageFilename(String string) {
        string = string.toLowerCase();
        for (String element : IMAGE_EXTENSIONS) {
            if (string.endsWith(element)) {
                return true;
            }
        }
        return false;
    }

    private static int parseIntent(Intent intent, String value, int defVal) {
        String str = intent.getStringExtra(value);
        if (str != null) {
            try {
                return Integer.parseInt(str);
            } catch (NumberFormatException ignored) {
            }
        }
        return defVal;
    }

    private static int clamp(int val, int min, int max) {
        return Math.min(Math.max(val, min), max);
    }

    private void setViewportLock(boolean wantLock) {
        if (mCurrentRenderer != null) {
            mCurrentRenderer.setViewportLock(wantLock);
        }
    }

    @Override
    public void previousFile() {
        switchToAnotherFile(-1);
    }

    @Override
    public void nextFile() {
        switchToAnotherFile(+1);
    }

    @Override
    public void deleteFile() {
        new AlertDialog.Builder(this).setMessage(getDeleteMessage())
                .setNegativeButton(android.R.string.cancel)
                .setPositiveButton(R.string.confirm_delete, (d, w) -> deleteActual())
                .show();
    }

    @Override
    public boolean allowsManageTags() {
        File file = new File(TAGS_CONFIG_FILE_LOCATION);
        if (file.exists()) {
            try {
                return mTagServiceConnector.get().isEnabled();
            } catch (RemoteException e) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "Error retrieving the IPC connector", e);
                }
            } catch (NullPointerException e) {
                Log.d(TAG, "IPC Tag Service Connector is null", e);
            }
        }
        return false;
    }

    @Override
    public boolean fileSupportsTags(Uri filePath) {
        if (filePath == null) {
            return false;
        }

        try {
            return mTagServiceConnector.get().isFileSupported(filePath.toString());
        } catch (RemoteException | NullPointerException e) {
            Log.e(TAG, "Failed to check if file supports tagging", e);
            return false;
        }
    }

    private String getDeleteMessage() {
        return mIsPhoto ? getString(R.string.photo_confirm_delete_message) :
                getString(R.string.default_confirm_delete_message);
    }

    private void deleteActual() {
        File myFile = getMyFile();

        try {
            Files.delete(myFile.toPath());
            finish();
        } catch (IOException e) {
            Log.e(TAG, "Failed to delete file " + myFile.getAbsolutePath(), e);
        }

        MediaScannerConnection.scanFile(this, new String[]{getMyFile().getAbsolutePath()}, null,
                (path, uri) -> Log.d(TAG, "onScanCompleted(" + path + ", " + uri + ")"));
    }

    private File[] getListOfFilesInMyDirectory() {
        final List<File> allFiles = new ArrayList<>(Arrays.asList(getMyFile().getParentFile().listFiles()));
        final List<File> filteredFiles = new ArrayList<>();
        for (File file : allFiles) {
            if (isSupportedFile(file)) {
                filteredFiles.add(file);
            }
        }
        filteredFiles.sort(mSortByDate ? BY_DATE_SORTER : BY_NAME_SORTER);
        return filteredFiles.toArray(new File[0]);
    }

    private int findMyFileIndex(File[] files) {
        final File myFile = getMyFile();
        final String myFilename = myFile.getName();
        for (int i = 0; i < files.length; i++) {
            if (myFilename.equalsIgnoreCase(files[i].getName())) {
                return i;
            }
        }
        throw new IllegalStateException(new FileNotFoundException("Did not find " + myFile.getAbsolutePath() + " in provided files"));
    }

    @NonNull
    private File getMyFile() {
        final String myPath = mFileLocation.getPath();
        if (myPath == null) {
            throw new IllegalStateException("No file location available");
        } else {
            return new File(myPath);
        }
    }

    private static String getMimeType(File file) {
        String mimeType = MimeTypeMap.getSingleton()
                .getMimeTypeFromExtension(MimeTypeMap.getFileExtensionFromUrl(file.toURI().toString().toLowerCase()));

        if (mimeType != null) {
            if (mimeType.startsWith("video/") || mimeType.startsWith("audio/")) {
                try {
                    final MediaMetadataRetriever retriever = new MediaMetadataRetriever();
                    retriever.setDataSource(file.getAbsolutePath());
                    mimeType = retriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_MIMETYPE);
                } catch (Exception ignored) {
                    // in case setDataSource() fails for corrupt file
                }
            }
        } else {
            MimeTypeMap mime = MimeTypeMap.getSingleton();
            int index = file.getName().lastIndexOf('.') + 1;
            String ext = file.getName().substring(index).toLowerCase();
            mimeType = mime.getMimeTypeFromExtension(ext);
        }

        if (mimeType != null && mimeType.equals("application/ogg")) {
            // for some reason audio files of type ogg get the wrong mime type
            mimeType = "audio/ogg";
        }

        return mimeType;
    }

    private void switchToAnotherFile(int indexIncrement) {
        final File[] files = getListOfFilesInMyDirectory();
        final int myIndex = findMyFileIndex(files);
        final int newIndex = myIndex + indexIncrement;
        if (isInRange(newIndex, 0, files.length - 1)) {
            File newFile = files[newIndex];

            // ScanFile returns a content://uri without needing to construct a custom provider for this app.
            MediaScannerConnection.scanFile(this,
                    new String[]{newFile.getAbsolutePath()}, null,
                    (path, uri) -> {
                        Intent newIntent = new Intent(Intent.ACTION_VIEW);
                        newIntent.addCategory(Intent.CATEGORY_DEFAULT);
                        newIntent.putExtra(SHOW_DELETE_EXTRA, mShowDelete);
                        newIntent.putExtra(SHOW_PREVIOUS_NEXT, mShowPrevNext);
                        newIntent.putExtra(SORT_BY_DATE, mSortByDate);
                        newIntent.putExtra(EXTRA_FILE_PATH, newFile.getAbsolutePath());
                        String mimeType = getMimeType(newFile);
                        if (mimeType != null) {
                            newIntent.setDataAndType(uri, mimeType);
                        } else {
                            newIntent.setData(uri);
                        }
                        newIntent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);

                        finish();
                        startActivity(newIntent);
                    });
        }
    }

    private boolean isFileIndexIncrementInRange(int indexIncrement) {
        final File[] files = getListOfFilesInMyDirectory();
        final int myIndex = findMyFileIndex(files);
        final int newIndex = myIndex + indexIncrement;
        return isInRange(newIndex, 0, files.length - 1);
    }

    @Override
    public void addBookmark(int pageNumber) {
        mDatabase.addBookmark(new Bookmark(mCurrentPage.get(), docLocation, docMod));
    }

    @Override
    public List<Bookmark> getAllBookmarks() {
        return mDatabase.getAllBookmarks(docLocation, docMod);
    }

    @Override
    public void deleteBookmark(int pageNumber) {
        mDatabase.deleteBookmark(new Bookmark(mCurrentPage.get(), docLocation, docMod));
    }

    public int getCurrentPage() {
        return mCurrentPage.get();
    }

    public int getPageCount() {
        return pageCount;
    }

    @Override
    public boolean canStoreBookmarks() {
        return mCanStoreBookmarks && docMod != null && docLocation != null;
    }

    private void handleTaggingServiceBeingConnected(ITagService service) {
        new Handler(Looper.getMainLooper()).post(() -> {
            try {
                if (service.isEnabled() && service.isFileSupported(mFileLocation.getPath())) {
                    for (ViewerFragment fragment : mViewerFragments) {
                        fragment.refreshCommands();
                    }
                }
            } catch (RemoteException e) {
                throw new IllegalStateException(e);
            }
        });
    }

    private void handleTaggingServiceDisconnected() {
        new Handler(Looper.getMainLooper()).post(() -> {
            for (ViewerFragment fragment : mViewerFragments) {
                fragment.refreshCommands();
            }
        });
    }

    /**
     * Launch the dialog that allow the user to select tags for the current file.
     */
    @SuppressLint("WrongConstant")
    public void manageTags() {
        try {
            final Intent intent = new Intent();
            intent.setClassName(ITagService.TAG_SERVICE_PACKAGE, ITagService.SELECTOR_ACTIVITY);
            intent.setData(mFileLocation);
            intent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);

            int pendingIntentFlag = Build.VERSION.SDK_INT >= Build.VERSION_CODES.S ? PendingIntent.FLAG_MUTABLE : 0;

            final PendingIntent pendingIntent = PendingIntent.getActivity(
                    this,
                    0,
                    intent,
                    pendingIntentFlag
            );
            mTagServiceConnector.get().showSelector(pendingIntent);
        } catch (RemoteException e) {
            Log.w(TAG, "Failed to show tag manager", e);
        }
    }

    private boolean isSupportedFile(File file) {
        return isImageFilename(file.getPath()) || file.getPath().toLowerCase().endsWith(DOCUMENT_EXTENSION);
    }

    @Override
    public void showSearchFragment() {
        showFragment(mSearchFragment, true);
        hideBookmarks();
        mCurrentRenderer = null;
    }

    @Override
    public void showDocument() {
        showDocumentFragment();
    }

    /**
     * Remove all stored search results.
     */
    @Override
    public void clearSearchResults() {
        mResultsByPage.clear();
        mDocumentViewFragment.clearHighlightRectangles();
    }

    /**
     * Add a search result to the stored results for display.
     *
     * @param searchResult The new search result.
     */
    @Override
    public void addToSearchResults(final PDFSearchHandler.SearchResult searchResult) {
        List<RectF> currentResults = mResultsByPage.get(searchResult.pageNumber);
        if (currentResults == null) {
            currentResults = new LinkedList<>();
            mResultsByPage.put(searchResult.pageNumber, currentResults);
        }
        currentResults.addAll(searchResult.highlightAreas);
    }

    /**
     * Returns whether or not a 'previous file' command should be shown.
     *
     * @return True if so, false if not.
     */
    @Override
    public boolean allowsPreviousFile() {
        return mShowPrevNext && isFileIndexIncrementInRange(-1);
    }

    /**
     * Returns whether or not a 'next file' command should be shown.
     *
     * @return True if so, false if not.
     */
    @Override
    public boolean allowsNextFile() {
        return mShowPrevNext && isFileIndexIncrementInRange(+1);
    }

    /**
     * Returns whether or not a 'delete file' command should be shown.
     *
     * @return True if so, false if not.
     */
    @Override
    public boolean allowsDeleteFile() {
        return mShowDelete;
    }

    /**
     * Exception thrown by the when files are unavailable.
     */
    public class FileUnavailableException extends Exception {
        /**
         * Constructor.
         *
         * @param explanation Details about the exception.
         */
        public FileUnavailableException(String explanation) {
            super(explanation);
        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode,
                                           @NonNull String[] permissions,
                                           @NonNull int[] grantResults) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults);

        if (PermissionUtils.INSTANCE.validatePermissionResult(requestCode, grantResults)) {
            onNewIntent(getIntent());
        } else {
            Toast.makeText(this, R.string.permission_error, Toast.LENGTH_LONG).show();
            finish();
        }
    }
}
