/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.adapters;

import android.annotation.TargetApi;
import android.os.Build;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.BaseAdapter;
import android.widget.TextView;

import com.realwear.documentviewer.R;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Consumer;

/**
 * Adapts a list of search terms into a view to appear on screen representing that recent search
 * term.
 */
@TargetApi(value = Build.VERSION_CODES.N)
public class RecentSearchesAdapter extends BaseAdapter {
    private final List<String> mCurrentTerms = new LinkedList<>();
    private Consumer<String> mOnClickHandler = null;

    @Override
    public int getCount() {
        return mCurrentTerms.size();
    }

    @Override
    public Object getItem(final int position) {
        return mCurrentTerms.get(position);
    }

    @Override
    public long getItemId(final int position) {
        return position;
    }

    @Override
    public View getView(final int position, final View convertView, final ViewGroup parent) {
        View outputView = convertView;
        if (outputView == null) {
            outputView =
                    LayoutInflater.from(parent.getContext()).inflate(R.layout.item_search_term,
                            parent, false);
        }
        ((TextView) outputView.findViewById(R.id.index)).setText(String.format("%d", position + 1));
        ((TextView) outputView.findViewById(R.id.search_text)).setText(mCurrentTerms.get(position));
        outputView.setOnClickListener(l -> {
            if (mOnClickHandler != null) {
                mOnClickHandler.accept(mCurrentTerms.get(position));
            }
        });
        return outputView;
    }

    /**
     * Updates the list of search terms to be the given collection - the input list is not modified.
     *
     * @param terms The new search terms - terms will be indexed in line with the collection's
     *              iteration order.
     */
    public void setCurrentTerms(Collection<String> terms) {
        mCurrentTerms.clear();
        mCurrentTerms.addAll(terms);
        notifyDataSetChanged();
    }

    /**
     * Sets a handler that is called when a search term's view is clicked/VC called.
     *
     * @param handler The handler for the click event - the string it will be given is the search
     *                term text clicked.
     */
    public void setOnClickHandler(Consumer<String> handler) {
        mOnClickHandler = handler;
    }
}
