/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.adapters;

import static com.realwear.internal.utils.Functional.runIfParameterNotNull;

import android.annotation.TargetApi;
import android.graphics.Bitmap;
import android.os.Build;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.BaseAdapter;
import android.widget.ImageView;
import android.widget.TextView;

import com.realwear.documentviewer.R;
import com.realwear.documentviewer.fragment.IDocumentController;
import com.realwear.documentviewer.memory.PDFSearchHandler;
import com.realwear.view.styles.ThemeAwareObject;
import com.realwear.view.styles.ThemeRegistrar;

import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;

/**
 * Adapts search results to display a view that shows the result on page and its page number.
 */
@TargetApi(value = Build.VERSION_CODES.N)
public class SearchResultsAdapter extends BaseAdapter {
    private final List<PDFSearchHandler.SearchResult> mResults = new CopyOnWriteArrayList<>();
    private final AtomicReference<IDocumentController> mController = new AtomicReference<>();
    private Function<PDFSearchHandler.SearchResult, Bitmap> mImageGenerator;

    @Override
    public int getCount() {
        return mResults.size();
    }

    @Override
    public Object getItem(final int position) {
        return mResults.get(position);
    }

    @Override
    public long getItemId(final int position) {
        return position;
    }

    @Override
    public View getView(final int position, final View convertView, final ViewGroup parent) {
        View viewOut = convertView;
        if (viewOut == null) {
            viewOut = LayoutInflater.from(parent.getContext()).inflate(R.layout.item_search_result,
                    parent, false);
            viewOut.findViewById(R.id.rounded_container).setClipToOutline(true);
        }
        final TextView gotoPage = viewOut.findViewById(R.id.goto_page_text);
        final PDFSearchHandler.SearchResult result = mResults.get(position);

        gotoPage.setText(viewOut.getContext().getString(R.string.select_page, result.pageNumber + 1));

        if (mImageGenerator != null) {
            final Bitmap imageOut = mImageGenerator.apply(result);
            ((ImageView) viewOut.findViewById(R.id.preview_display)).setImageBitmap(imageOut);
        }

        viewOut.setOnClickListener(v -> executeGotoPage(result.pageNumber));
        registerAllThemedChildrenNow(viewOut);
        return viewOut;
    }

    /**
     * Adds a search result to the adapter.
     *
     * @param result The new search result to add.
     */
    public void addItem(PDFSearchHandler.SearchResult result) {
        mResults.add(result);
        notifyDataSetChanged();
        runIfParameterNotNull(controller -> controller.addToSearchResults(result), mController.get());
    }

    /**
     * Sets the document controller to allow the 'go to page' command to be executed.
     *
     * @param newController The new controller.
     */
    public void setController(IDocumentController newController) {
        mController.set(newController);
    }

    /**
     * Clears all search results from the adapter.
     */
    public void clearItems() {
        mResults.clear();
        notifyDataSetChanged();
    }

    /**
     * Execute a 'go to page' command with the given page number (zero indexed).
     *
     * @param pageNumber The (zero indexed) page number to go to. (i.e. page 1 is 0, page 2 is 1,
     *                   etc).
     */
    public void executeGotoPage(int pageNumber) {
        runIfParameterNotNull(controller -> {
            controller.showDocument();
            controller.gotoPage(pageNumber);
        }, mController.get());
    }

    /**
     * Sets the function to generate a bitmap representing a given search result.
     *
     * @param newImageFunction The new image function.
     */
    public void setImageGenerator(Function<PDFSearchHandler.SearchResult, Bitmap> newImageFunction) {
        mImageGenerator = newImageFunction;
    }

    /**
     * Performs an instant registration of any Theme Aware views or child views to ensure they
     * are properly themed before being shown to the end user.
     *
     * @param view The view to search through.
     */
    private void registerAllThemedChildrenNow(View view) {
        if (view instanceof ViewGroup) {
            for (int i = 0; i < ((ViewGroup) view).getChildCount(); i++) {
                registerAllThemedChildrenNow(((ViewGroup) view).getChildAt(i));
            }
        }
        if (view instanceof ThemeAwareObject) {
            runIfParameterNotNull(registrar -> registrar.ensureObjectRegistered((ThemeAwareObject) view),
                    ThemeRegistrar.getInstance());
        }
    }
}
