/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.fragment;

import android.app.Fragment;
import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.ImageView;

import androidx.annotation.Nullable;

import com.realwear.documentviewer.R;
import com.realwear.view.StyledTextView;
import com.realwear.view.commandbar.Action;
import com.realwear.view.commandbar.CommandBar;

/**
 * Fragment that displays a loading screen while a document is rendering.
 */
public class DocumentLoadingFragment extends Fragment {
    StyledTextView mFileName;
    ImageView mThumbnail;
    CommandBar mCommandBar;

    private ViewerFragment.OnFragmentLoadedListener mOnFragmentLoadedListener = null;

    /**
     * Create a new document loading fragment instance.
     *
     * @return The new instance.
     */
    public static DocumentLoadingFragment newInstance() {
        return new DocumentLoadingFragment();
    }

    @Nullable
    @Override
    public View onCreateView(
            LayoutInflater inflater, ViewGroup container, Bundle savedInstanceState) {
        return inflater.inflate(R.layout.fragment_document_loading, container, false);
    }

    @Override
    public void onViewCreated(View view, @Nullable Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);
        mFileName = view.findViewById(R.id.load_filename);
        mThumbnail = view.findViewById(R.id.load_thumbnail);
        mCommandBar = view.findViewById(R.id.command_bar);

        Action cancelLoad = new Action(this::cancelFileLoad, getString(R.string.cancel));
        mCommandBar.addAction(cancelLoad);
    }

    @Override
    public void onResume() {
        super.onResume();
        mCommandBar.onResume();

        //
        // The renderer is still setting up in during the onResume part of the life cycle, so defer
        // any processing.
        //
        new Handler(Looper.getMainLooper()).post(() -> {
            if (mOnFragmentLoadedListener != null) {
                mOnFragmentLoadedListener.onFragmentLoaded();
            }
        });
    }

    @Override
    public void onPause() {
        mCommandBar.onPause();
        super.onPause();
    }

    /**
     * Set the filename to be displayed while the file loads.
     *
     * @param filename The name of the file being loaded.
     */
    public void setFilename(String filename) {
        if (filename != null) {
            mFileName.setText(filename);
        }
    }

    /**
     * Called when the user decides to cancel file load.
     */
    public void cancelFileLoad() {
        getActivity().finish();
    }

    /**
     * Set the callback to be invoked when the fragment is loaded.
     *
     * @param listener The listener to invoke.
     */
    public void setFragmentLoadedListener(ViewerFragment.OnFragmentLoadedListener listener) {
        mOnFragmentLoadedListener = listener;
    }
}
