/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.fragment;

import static com.realwear.documentviewer.memory.CommandRegister.CommandInfo;

import android.annotation.TargetApi;
import android.graphics.RectF;
import android.os.Build;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.Toast;

import androidx.annotation.Nullable;

import com.core.realwear.sdk.util.WearToast;
import com.realwear.documentviewer.MainActivity;
import com.realwear.documentviewer.R;
import com.realwear.documentviewer.object.Bookmark;
import com.realwear.view.commandbar.Action;
import com.realwear.view.commandbar.pagedisplay.PageDisplay;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;

/**
 * Represents a PDF document view.
 */
@TargetApi(value = Build.VERSION_CODES.M)
public class DocumentViewerFragment extends ViewerFragment {
    private PageDisplay mPageDisplay;

    public static DocumentViewerFragment newInstance() {
        return new DocumentViewerFragment();
    }

    @Nullable
    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container, Bundle savedInstanceState) {
        return inflater.inflate(R.layout.fragment_documentviewer, container, false);
    }

    @Override
    public void onViewCreated(View view, @Nullable Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);

        mPageDisplay = view.findViewById(R.id.page_display);
        mPageDisplay.setVisibility(View.VISIBLE);

        mCommandBar.addAction(new Action(() -> {
                    mCommandBar.hideMoreOptions();
                    if (mFragmentController != null) {
                        mFragmentController.showBookmarkFragment();
                    }
                }, getString(R.string.my_booksmarks), getContext().getDrawable(R.drawable.ic_bookmarks)),
                false);
        getMainActivity().currentPage.addListener(pageNumber -> mPageDisplay.setPageNumber(pageNumber + 1));
    }

    /**
     * Clear all search results highlighting.
     */
    public void clearHighlightRectangles() {
        if (mRenderer != null) {
            mRenderer.clearHighlightRectangles();
        }
    }

    /**
     * Add a list of rectangles to highlight for search results.
     *
     * @param rectangles The rectangles to add.
     */
    public void addHighlightRectangles(List<RectF> rectangles) {
        if (mRenderer != null) {
            mRenderer.addHighlightRectangles(rectangles);
        }
    }

    @Override
    protected Collection<CommandInfo> getHiddenCommands() {
        final List<CommandInfo> commands = new LinkedList<>();
        if (canBookMark()) {
            if (isPageBookmarked()) {
                commands.add(new CommandInfo(getString(R.string.delete_bookmark),
                        () -> {
                            mFragmentController.deleteBookmark(getCurrentPage());
                            WearToast.makeWearText(getActivity(), getString(R.string.bookmark_deleted_currentpage), Toast.LENGTH_LONG).show();
                            refreshCommands();
                        }));
            } else {
                commands.add(new CommandInfo(getString(R.string.bookmark_currentpage),
                        () -> {
                            mFragmentController.addBookmark(getCurrentPage());
                            WearToast.makeWearText(getActivity(), getString(R.string.bookmark_add_currentpage), Toast.LENGTH_LONG).show();
                            refreshCommands();
                        }));
            }
        }
        if (getCurrentPage() > 0) {
            commands.add(new CommandInfo(getString(R.string.previousPage),
                    () -> getMainActivity().gotoPrevPage()));
        }
        if (getCurrentPage() + 1 != getMainActivity().getPageCount()) {
            commands.add(new CommandInfo(getString(R.string.nextPage),
                    () -> getMainActivity().gotoNextPage()));
        }
        for (int i = 1; i <= getPageCount(); i++) {
            final int page = i;
            commands.add(new CommandInfo(getString(R.string.select_page, page),
                    () -> {
                        getMainActivity().gotoPage(page - 1);
                        mZoomControl.selectNewLevel(1);
                    },
                    false));
        }
        mHiddenCommandRegistrar.addToHelp(getString(R.string.select_page_range, 1,
                getPageCount()));
        mPageDisplay.setMaximumPage(getPageCount());

        return commands;
    }

    /**
     * Returns if bookmarks can be stored.
     *
     * @return True if possible.
     */
    private boolean canBookMark() {
        return mFragmentController.canStoreBookmarks();
    }

    /**
     * Returns whether or not the current page is bookmarked.
     *
     * @return True if it is, else false.
     */
    private boolean isPageBookmarked() {
        for (Bookmark bookmark : getMainActivity().getAllBookmarks()) {
            if (bookmark.getBookmarkNumber() == getCurrentPage()) {
                return true;
            }
        }

        return false;
    }

    @Override
    public View getFragmentView() {
        return getView();
    }

    /**
     * Returns the current page number.
     *
     * @return A zero-starting page number.
     */
    private int getCurrentPage() {
        return getMainActivity().getCurrentPage();
    }

    /**
     * Returns the current page count.
     *
     * @return The total page count.
     */
    private int getPageCount() {
        return getMainActivity().getPageCount();
    }

    /**
     * Returns the parent activity.
     *
     * @return The parent activity.
     */
    private MainActivity getMainActivity() {
        return Objects.requireNonNull((MainActivity) getActivity());
    }
}
