/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.fragment;

import static com.realwear.internal.utils.Functional.runIfParameterNotNull;
import static java.util.stream.Collectors.joining;

import android.annotation.TargetApi;
import android.app.Fragment;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.res.ColorStateList;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.drawable.GradientDrawable;
import android.graphics.pdf.PdfRenderer;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.inputmethod.InputMethodManager;
import android.widget.EditText;
import android.widget.GridView;
import android.widget.ImageView;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.realwear.documentviewer.R;
import com.realwear.documentviewer.adapters.RecentSearchesAdapter;
import com.realwear.documentviewer.adapters.SearchResultsAdapter;
import com.realwear.documentviewer.memory.PDFSearchHandler;
import com.realwear.documentviewer.memory.RecentSearches;
import com.realwear.documentviewer.views.BitmapDisplay;
import com.realwear.internal.utils.Observable;
import com.realwear.view.commandbar.Action;
import com.realwear.view.commandbar.CommandBar;
import com.realwear.view.styles.Theme;
import com.realwear.view.styles.ThemeAwareObject;
import com.realwear.view.styles.ThemeRegistrar;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

/**
 * Represents a search file UX.
 */
@TargetApi(value = Build.VERSION_CODES.N)
public class SearchFragment extends Fragment implements ThemeAwareObject {
    private static final String ACTION_SPEECH_EVENT = "com.realwear.wearhf.intent.action.SPEECH_EVENT";
    private static final String EXTRA_RESULT = "command";

    private enum UxMode {
        NONE,
        RECENT,
        RESULTS
    }

    private final Handler mHandler = new Handler(Looper.getMainLooper());
    private final Observable<Boolean> mKeyboardIsVisible = new Observable<>(Boolean.FALSE);
    private EditText mEditTextView;
    private PDFSearchHandler mSearcher = null;
    private IDocumentController mController;
    private final SearchResultsAdapter mAdapter = new SearchResultsAdapter();
    private final RecentSearchesAdapter mRecentSearchesAdapter = new RecentSearchesAdapter();
    private final Observable<Integer> mBackgroundColour = new Observable<>(0);
    private final Observable<Integer> mSearchTextColour = new Observable<>(0);
    private final Set<Integer> mPagesToGoto = new CopyOnWriteArraySet<>();
    private View mResultsArea;
    private View mBusyIcon;
    private View mRecentSearchesArea;
    private ImageView mFadeOutView;
    private RecentSearches mSearches;
    private CommandBar mCommandBar;
    private Action mActionRecentSearches;
    private TextView mResultCounter;
    private GridView mResultsGrid;
    private int mResultsCount = 0;
    private final Runnable mOnSearchFinished = this::handleSearchFinished;
    private final List<String> mCurrentTerms = new LinkedList<>();
    private UxMode mUxMode = UxMode.NONE;

    public static SearchFragment newInstance() {
        return new SearchFragment();
    }

    @Nullable
    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container, Bundle savedInstanceState) {
        return inflater.inflate(R.layout.fragment_search, container, false);
    }

    @Override
    public void onViewCreated(View view, @Nullable Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);

        mEditTextView = view.findViewById(R.id.search_term);

        mBackgroundColour.addListener(colour -> mEditTextView.setBackgroundTintList(ColorStateList.valueOf(colour)));
        mSearchTextColour.addListener(colour -> mEditTextView.setTextColor(colour));
        mEditTextView.setBackgroundTintList(ColorStateList.valueOf(mBackgroundColour.get()));

        view.getViewTreeObserver().addOnGlobalLayoutListener(() -> {
            Rect measureRect = new Rect(); //you should cache this, onGlobalLayout can get called often
            view.getWindowVisibleDisplayFrame(measureRect);
            // measureRect.bottom is the position above soft keypad
            int keypadHeight = view.getRootView().getHeight() - measureRect.bottom;

            if (keypadHeight > 0) {
                // keyboard is opened
                mHandler.post(() -> mKeyboardIsVisible.update(Boolean.TRUE));
            } else {
                //store keyboard state to use in onBackPress if you need to
                mHandler.post(() -> mKeyboardIsVisible.update(Boolean.FALSE));
            }
        });

        mKeyboardIsVisible.addListener(isVisible -> {
            if (Boolean.FALSE.equals(isVisible)) {
                mEditTextView.clearFocus();
                handleNewSearchTerm();
            }
        });
        mResultsGrid = view.findViewById(R.id.results_view);
        mResultsGrid.setAdapter(mAdapter);
        mResultsArea = view.findViewById(R.id.results_area);

        mSearches = new RecentSearches(view.getContext());
        final GridView recentGrid = view.findViewById(R.id.recent_searches_grid);
        recentGrid.setAdapter(mRecentSearchesAdapter);
        mRecentSearchesAdapter.setCurrentTerms(mSearches.getRecentSearches());

        mRecentSearchesArea = view.findViewById(R.id.recent_searches_view);

        mBusyIcon = view.findViewById(R.id.busy_icon);
        mBusyIcon.setVisibility(View.INVISIBLE);

        mRecentSearchesAdapter.setOnClickHandler(searchTerm -> {
            mEditTextView.setText(searchTerm);
            startSearch(searchTerm);
        });

        mFadeOutView = view.findViewById(R.id.fade_view);
        mCommandBar = view.findViewById(R.id.command_bar);
        mResultCounter = view.findViewById(R.id.results_count);
        final Action actionNewSearch = new Action(() -> {
            mSearcher.cancelSearch();
            mEditTextView.setText("");
            if (mEditTextView.requestFocus()) {
                InputMethodManager inputMethodManager =
                        getContext().getSystemService(InputMethodManager.class);
                inputMethodManager.showSoftInput(mEditTextView, InputMethodManager.SHOW_IMPLICIT);
            }
        }, getString(R.string.new_search));
        mActionRecentSearches = new Action(this::showRecentSearches, getString(R.string.show_recent));
        mCommandBar.addAction(actionNewSearch, mActionRecentSearches);

        if (mEditTextView.requestFocus()) {
            InputMethodManager inputMethodManager =
                    getContext().getSystemService(InputMethodManager.class);
            inputMethodManager.showSoftInput(mEditTextView, InputMethodManager.SHOW_IMPLICIT);
        }
        mRecentSearchesArea.setVisibility(View.INVISIBLE);
        mResultsArea.setVisibility(View.INVISIBLE);
        mUxMode = UxMode.NONE;

        runIfParameterNotNull(r -> r.ensureObjectRegisteredLater(this),
                ThemeRegistrar.getInstance());
    }

    /**
     * Set's the PDF searcher object.
     *
     * @param searcher The object.
     */
    public void setSearcher(PDFSearchHandler searcher) {
        mSearcher = searcher;
    }

    /**
     * Set's the PDF renderer object for rendering results.
     *
     * @param renderer The object.
     */
    public void setRenderer(PdfRenderer renderer) {
        mAdapter.setImageGenerator(result -> {
            final int outputWidth =
                    getResources().getDimensionPixelSize(R.dimen.search_preview_width);
            final int outputHeight =
                    getResources().getDimensionPixelSize(R.dimen.search_preview_height);
            final PdfRenderer.Page page = renderer.openPage(result.pageNumber);

            Bitmap baseBitmap = Bitmap.createBitmap(page.getWidth(), page.getHeight(),
                    Bitmap.Config.ARGB_8888);
            baseBitmap.eraseColor(Color.WHITE);

            page.render(baseBitmap, null, null,
                    PdfRenderer.Page.RENDER_MODE_FOR_DISPLAY);
            page.close();

            Canvas canvas = new Canvas(baseBitmap);

            RectF renderRectF = new RectF(result.highlightAreas.get(0));
            final Paint highlightPaint = new Paint();
            highlightPaint.setColor(BitmapDisplay.HIGHLIGHT_COLOUR);
            for (RectF r : result.highlightAreas) {
                renderRectF.union(r);
                canvas.drawRect(new RectF(r.left * page.getWidth(),
                        r.top * page.getHeight(),
                        r.right * page.getWidth(),
                        r.bottom * page.getHeight()
                ), highlightPaint);
            }
            final int centreX = (int) (renderRectF.centerX() * page.getWidth());
            final int centreY = (int) (renderRectF.centerY() * page.getHeight());
            final int left = centreX - (outputWidth / 2);
            final int top = centreY - (outputHeight / 2);

            return Bitmap.createBitmap(baseBitmap,
                    limitToRange(0, left, page.getWidth() - outputWidth),
                    limitToRange(0, top, page.getHeight() - outputHeight),
                    outputWidth,
                    outputHeight);
        });
    }

    /**
     * Sets the document controller object.
     *
     * @param newController The object.
     */
    public void setController(IDocumentController newController) {
        mController = newController;
        mAdapter.setController(newController);
    }

    /**
     * Handle new search term being entered.
     */
    private void handleNewSearchTerm() {
        final String searchTerm = mEditTextView.getText().toString();
        if (searchTerm.trim().isEmpty()) {
            showRecentSearches();
        } else {
            startSearch(searchTerm);
        }
    }

    /**
     * Starts a search; assumes the string is not null or empty.
     *
     * @param searchTerm The text to search for.
     */
    private void startSearch(@NonNull final String searchTerm) {
        mUxMode = UxMode.RESULTS;
        if (mController != null) {
            mController.clearSearchResults();
        }
        resetCounterNumber();
        mCommandBar.showAction(mActionRecentSearches);
        mSearches.addRecentSearch(searchTerm);
        mPagesToGoto.clear();
        mAdapter.clearItems();
        mRecentSearchesArea.setVisibility(View.INVISIBLE);
        mResultsArea.setVisibility(View.VISIBLE);
        mBusyIcon.setVisibility(View.VISIBLE);
        if (mSearcher != null) {
            mSearcher.beginSearch(searchTerm,
                    result -> {
                        mAdapter.addItem(result);
                        incrementCounterNumber();
                        if (mPagesToGoto.add(result.pageNumber)) {
                            mResultsGrid.setContentDescription(
                                    "hf_add_commands:" +
                                            mPagesToGoto.stream()
                                                    .map(n -> getString(R.string.select_page, n + 1))
                                                    .collect(joining("|"))
                            );
                        }
                    });
            mSearcher.errorFlag.addListener(mOnSearchFinished);
            mSearcher.finishFlag.addListener(mOnSearchFinished);
        }
    }

    @Override
    public void handleThemeUpdate(final Theme theme) {
        mBackgroundColour.update(theme.getColor(Theme.ColorType.BACKGROUND));
        mSearchTextColour.update(theme.getColor(Theme.ColorType.ON_BACKGROUND));

        int colour = theme.getColor(Theme.ColorType.SURFACE);
        GradientDrawable gradientDrawable = (GradientDrawable) mFadeOutView.getDrawable();
        gradientDrawable.setColors(new int[]{
                Color.argb(0, Color.red(colour), Color.green(colour), Color.blue(colour)),
                colour
        });
    }

    @Override
    public void onHiddenChanged(final boolean hidden) {
        super.onHiddenChanged(hidden);
        if (hidden) {
            mSearcher.cancelSearch();
        } else {
            if (mEditTextView.requestFocus()) {
                InputMethodManager inputMethodManager =
                        getContext().getSystemService(InputMethodManager.class);
                inputMethodManager.showSoftInput(mEditTextView, InputMethodManager.SHOW_IMPLICIT);
            }
            mRecentSearchesArea.setVisibility(View.INVISIBLE);
            mResultsArea.setVisibility(View.INVISIBLE);
        }
    }

    @Override
    public void onResume() {
        super.onResume();
        mCommandBar.onResume();
        getContext().registerReceiver(mCommandReceiver, new IntentFilter(ACTION_SPEECH_EVENT));

    }

    @Override
    public void onPause() {
        getContext().unregisterReceiver(mCommandReceiver);
        mCommandBar.onPause();
        super.onPause();
    }

    /**
     * Limit a value to a range
     *
     * @param min   The minimum value.
     * @param value The input value.
     * @param max   The maximum value.
     * @return The limited value.
     */
    private int limitToRange(int min, int value, int max) {
        return Math.max(min, Math.min(value, max));
    }

    /**
     * Called when a search has been completed to update the UX.
     */
    private void handleSearchFinished() {
        mHandler.post(() -> {
            mBusyIcon.setVisibility(View.INVISIBLE);
            mSearcher.errorFlag.removeListener(mOnSearchFinished);
            mSearcher.finishFlag.removeListener(mOnSearchFinished);
        });
    }

    private final BroadcastReceiver mCommandReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(final Context context, final Intent intent) {
            if (intent == null || intent.getAction() == null) {
                return;
            }
            if (intent.getAction().equals(ACTION_SPEECH_EVENT)) {
                final String command = intent.getStringExtra(EXTRA_RESULT);
                if (command == null) {
                    return;
                }
                if (mUxMode == UxMode.RESULTS) {
                    for (final int n : mPagesToGoto) {
                        if (command.equalsIgnoreCase(getString(R.string.select_page, n + 1))) {
                            mAdapter.executeGotoPage(n);
                            return;
                        }
                    }
                } else if (mUxMode == UxMode.RECENT) {
                    for (final String term : mCurrentTerms) {
                        if (command.equalsIgnoreCase(term)) {
                            mEditTextView.setText(command);
                            startSearch(command);
                            return;
                        }
                    }
                }
            }
        }
    };

    /**
     * Shows a UX of recent searches.
     */
    private void showRecentSearches() {
        mSearcher.cancelSearch();
        mUxMode = UxMode.RECENT;
        mCommandBar.hideAction(mActionRecentSearches);
        mRecentSearchesArea.setVisibility(View.VISIBLE);
        mResultsArea.setVisibility(View.INVISIBLE);
        mBusyIcon.setVisibility(View.INVISIBLE);
        mCurrentTerms.clear();
        mCurrentTerms.addAll(mSearches.getRecentSearches());
        mRecentSearchesAdapter.setCurrentTerms(mCurrentTerms);
        mRecentSearchesArea.setContentDescription("hf_add_commands:" +
                mCurrentTerms.stream().map(term -> "#" + term).collect(joining("|")));
    }

    /**
     * Reset's the on screen results counter to zero.
     */
    private void resetCounterNumber() {
        mResultsCount = 0;
        updateCounterDisplay();
    }

    /**
     * Increments the current results counter and updates the display.
     */
    private void incrementCounterNumber() {
        mResultsCount += 1;
        updateCounterDisplay();
    }

    /**
     * Updates the results counter display.
     */
    private void updateCounterDisplay() {
        mResultCounter.setText(
                String.format(
                        getResources().getQuantityText(R.plurals.results_count, mResultsCount).toString(),
                        mResultsCount));
    }
}
