/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.fragment;

import android.app.Fragment;
import android.graphics.Bitmap;
import android.net.Uri;
import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.view.View;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.lifecycle.Lifecycle;
import androidx.lifecycle.LifecycleOwner;
import androidx.lifecycle.LifecycleRegistry;

import com.realwear.documentviewer.R;
import com.realwear.documentviewer.memory.CommandRegister;
import com.realwear.documentviewer.memory.CommandRegister.CommandInfo;
import com.realwear.documentviewer.object.IRenderer;
import com.realwear.documentviewer.views.BitmapDisplay;
import com.realwear.documentviewer.views.DisplayPreview;
import com.realwear.view.LevelControl;
import com.realwear.view.commandbar.Action;
import com.realwear.view.commandbar.CommandBar;
import com.realwear.views.listcontrol.headtracker.HFHeadtrackerListener;
import com.realwear.views.listcontrol.headtracker.HFHeadtrackerManager;
import com.realwear.views.listcontrol.headtracker.HFHeadtrackerManagerRotation;

import java.util.Collection;

public abstract class ViewerFragment extends Fragment implements HFHeadtrackerListener,
        IRenderer, LifecycleOwner {

    private final HFHeadtrackerManager mHeadTracker = new HFHeadtrackerManagerRotation(this);
    protected CommandBar mCommandBar;
    @Nullable
    protected BitmapDisplay mRenderer;
    private DisplayPreview mPreviewRender;
    private final LifecycleRegistry mLifecycle = new LifecycleRegistry(this);
    protected CommandRegister mHiddenCommandRegistrar;
    private View mHiddenCommandsArea;
    protected LevelControl mZoomControl;
    private TextView mFilenameTextView;
    private final Handler mHandler = new Handler(Looper.getMainLooper());
    private Action mActionFreeze, mActionControl, mActionManageTags, mActionDelete,
            mActionPrevious, mActionNext;
    protected IFragmentController mFragmentController;
    private int mInitialZoom = 0;

    private Uri mFile;
    private String mFilename;

    private OnFragmentLoadedListener mOnFragmentLoadedListener = null;

    @Override
    public void onResume() {
        super.onResume();
        mLifecycle.handleLifecycleEvent(Lifecycle.Event.ON_RESUME);
        if (mCommandBar != null) {
            mCommandBar.onResume();
        }
        mHeadTracker.startHeadtracker(
                getActivity(), HFHeadtrackerManager.TrackingMode.TRACK_ALL_MOTION);

        //
        // The renderer is still setting up in during the onResume part of the life cycle, so defer
        // any processing.
        //
        new Handler(Looper.getMainLooper()).post(() -> {
            if (mOnFragmentLoadedListener != null) {
                mOnFragmentLoadedListener.onFragmentLoaded();
            }
        });
    }

    @Override
    public void onPause() {
        mHeadTracker.stopHeadtracker();
        if (mCommandBar != null) {
            mCommandBar.onPause();
        }
        mLifecycle.handleLifecycleEvent(Lifecycle.Event.ON_PAUSE);
        super.onPause();
    }

    @Override
    public void onViewCreated(View view, @Nullable Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);

        view.findViewById(R.id.page_display).setVisibility(View.GONE);

        mHiddenCommandRegistrar = new CommandRegister(view.getContext());
        mCommandBar = view.findViewById(R.id.command_bar);
        mPreviewRender = view.findViewById(R.id.preview);
        mHiddenCommandsArea = view.findViewById(R.id.hidden_commands_area);
        mRenderer = view.findViewById(R.id.renderer);
        mRenderer.setDisplayPreview(mPreviewRender);
        mZoomControl = view.findViewById(R.id.zoom_control);
        mZoomControl.level.addListener(this::setZoomLevel);

        mFilenameTextView = view.findViewById(R.id.filename_text);
        mFilenameTextView.setText(mFilename);

        mZoomControl.level.addListener(level -> {
            if (level > 1) {
                mFilenameTextView.setVisibility(View.GONE);
            } else {
                mFilenameTextView.setVisibility(View.VISIBLE);
            }
        }, true);

        mActionFreeze = new Action(() -> setViewportLock(true), getString(R.string.freeze_document));
        mActionControl = new Action(() -> setViewportLock(false), getString(R.string.control_document));
        mActionDelete = new Action(() -> mFragmentController.deleteFile(), getString(R.string.delete_file));
        mActionPrevious = new Action(() -> mFragmentController.previousFile(),
                getString(R.string.previous_file));
        mActionNext = new Action(() -> mFragmentController.nextFile(), getString(R.string.next_file));
        mActionManageTags = new Action(() -> {
            if (mFragmentController != null) {
                mFragmentController.manageTags();
            }
        }, getString(R.string.manage_tags));

        mCommandBar.addAction(mActionControl, mActionFreeze, mActionManageTags, mActionDelete,
                mActionPrevious, mActionNext);
        mCommandBar.hideAction(mActionControl, mActionFreeze, mActionManageTags, mActionDelete,
                mActionPrevious, mActionNext);

        mRenderer.setZoomLevel(mInitialZoom);

        mHandler.postDelayed(this::refreshCommands, 50);

        mLifecycle.addObserver(mHiddenCommandRegistrar);
        mLifecycle.addObserver(mRenderer);
        mLifecycle.handleLifecycleEvent(Lifecycle.Event.ON_CREATE);
    }

    @Override
    public void setBitmap(Bitmap bitmap) {
        if (mRenderer != null) {
            mRenderer.setBitmap(bitmap);
            mPreviewRender.setPreviewBitmap(bitmap);
            refreshCommands();
        }
    }

    @Override
    public void setZoomLevel(int zoomLevel) {
        if (mRenderer != null) {
            mRenderer.setZoomLevel(zoomLevel - 1);
            refreshCommands();
        } else {
            mInitialZoom = zoomLevel - 1;
        }
    }

    @NonNull
    @Override
    public Lifecycle getLifecycle() {
        return mLifecycle;
    }

    @Override
    public boolean isDocumentLocked() {
        return mRenderer.isViewLocked();
    }

    /**
     * Set the callback to be invoked when the fragment is loaded.
     *
     * @param listener The listener to invoke.
     */
    public void setFragmentLoadedListener(OnFragmentLoadedListener listener) {
        mOnFragmentLoadedListener = listener;
    }

    /**
     * Refresh the UX commands (command bar and hidden commands).
     */
    public void refreshCommands() {
        if (mHiddenCommandRegistrar == null) {
            return;
        }
        mHiddenCommandRegistrar.clear();
        for (CommandInfo info : getHiddenCommands()) {
            mHiddenCommandRegistrar.addCommand(info);
        }
        if (mRenderer.getZoomLevel() > 0) {
            mHiddenCommandRegistrar.addCommand(getString(R.string.comands_reset), () -> resetDocument(true));
        }
        final String newDirective = mHiddenCommandRegistrar.getCommandsDirective();
        final CharSequence oldDirective = mHiddenCommandsArea.getContentDescription();
        if (oldDirective == null || !newDirective.equals(oldDirective.toString())) {
            mHiddenCommandsArea.setContentDescription(mHiddenCommandRegistrar.getCommandsDirective());
        }
        if (mRenderer.getZoomLevel() > 0) {
            if (mRenderer.isViewLocked()) {
                mCommandBar.showAction(mActionControl);
                mCommandBar.hideAction(mActionFreeze);
            } else {
                mCommandBar.showAction(mActionFreeze);
                mCommandBar.hideAction(mActionControl);
            }
        } else {
            mCommandBar.hideAction(mActionControl, mActionFreeze);
        }

        if (mFragmentController != null) {
            mCommandBar.setActionVisible(mFragmentController.allowsManageTags(), mActionManageTags);
            mCommandBar.setActionVisible(mFragmentController.allowsDeleteFile(), mActionDelete);
            mCommandBar.setActionVisible(mFragmentController.allowsNextFile(), mActionNext);
            mCommandBar.setActionVisible(mFragmentController.allowsPreviousFile(), mActionPrevious);

            if (!mFragmentController.fileSupportsTags(mFile)) {
                mCommandBar.hideAction(mActionManageTags);
            }
        }
    }

    /**
     * Implemented by children to return the currently available hidden commands list.
     *
     * @return A collection of hidden commands to recognize.
     */
    protected abstract Collection<CommandInfo> getHiddenCommands();

    @Override
    public void setViewportLock(boolean isLocked) {
        if (mRenderer != null) {
            mRenderer.lockView(isLocked);
            refreshCommands();
        }
    }

    @Override
    public void resetDocument(boolean isLocked) {
        mRenderer.resetView(isLocked, true);
    }

    @Override
    public void onHeadMoved(float deltaX, float deltaY) {
        mRenderer.translateView(deltaX, deltaY);
    }

    @Override
    public void gotoLocation(int x, int y) {
        mRenderer.setLocation(x, y);
    }

    @Override
    public View getFragmentView() {
        return getView();
    }

    /**
     * Sets the file that's being displayed.
     *
     * @param file     The file that is being displayed.
     * @param filename The name of the file that is being displayed.
     */
    public void setFile(Uri file, String filename) {
        mFile = file;
        mFilename = filename;

        if (mFilenameTextView != null) {
            mFilenameTextView.setText(filename);
        }
    }

    /**
     * Set's the fragment controller object.
     *
     * @param controller The object.
     */
    public void setFragmentController(IFragmentController controller) {
        mFragmentController = controller;
    }

    /**
     * Interface definition for a callback to be invoked when the fragment is loaded.
     */
    public interface OnFragmentLoadedListener {
        /**
         * Called when the fragment is loaded.
         */
        void onFragmentLoaded();
    }
}
