/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.memory;

import android.annotation.TargetApi;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Build;

import androidx.lifecycle.Lifecycle;
import androidx.lifecycle.LifecycleObserver;
import androidx.lifecycle.OnLifecycleEvent;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

/**
 * Stores and dispatches action to execute received speech commands. Handles registering for
 * speech commands if registered as a lifecycle observer.
 */
@TargetApi(value = Build.VERSION_CODES.O_MR1)
public class CommandRegister implements LifecycleObserver {
    public static final String ACTION_SPEECH_EVENT = "com.realwear.wearhf.intent.action" +
            ".SPEECH_EVENT";
    public static final String EXTRA_COMMAND = "com.realwear.wearhf.intent.extra.COMMAND";

    private final Context mContext;
    private final Map<String, Runnable> mCommandToActionMap = new HashMap<>();
    private final StringBuilder mHelpString = new StringBuilder();

    public static class CommandInfo {
        /**
         * Constructor.
         *
         * @param command The command's text.
         * @param action  The runnable to execute on the command.
         */
        public CommandInfo(final String command, final Runnable action) {
            this.command = command;
            this.action = action;
            this.addToHelp = true;
        }

        /**
         * Constructor.
         *
         * @param command   The command's text.
         * @param action    The runnable to execute on the command.
         * @param addToHelp True if the command should show up on 'show help'.
         */
        public CommandInfo(final String command, final Runnable action, boolean addToHelp) {
            this.command = command;
            this.action = action;
            this.addToHelp = addToHelp;
        }

        public final String command;
        public final Runnable action;
        public final boolean addToHelp;
    }

    /**
     * Constructor.
     *
     * @param context Operating context.
     */
    public CommandRegister(Context context) {
        mContext = context;
    }

    /**
     * Add a command to the registry with its associated action.
     *
     * @param command The text of the command.
     * @param action  The action to execute if the command is received.
     */
    public void addCommand(String command, Runnable action) {
        addCommand(command, action, true);
    }

    /**
     * Add a command to the registry with its associated action.
     *
     * @param command    The text of the command.
     * @param action     The action to execute if the command is received.
     * @param showOnHelp True if the command should be added to 'show help'.
     */
    public void addCommand(String command, Runnable action, boolean showOnHelp) {
        mCommandToActionMap.put(command, action);
        if (showOnHelp) {
            addToHelp(command);
        }
    }

    /**
     * Add a command to the registry with its associated action.
     *
     * @param commandInformation The information about the command.
     */
    public void addCommand(CommandInfo commandInformation) {
        addCommand(commandInformation.command, commandInformation.action, commandInformation.addToHelp);
    }

    /**
     * Remove all commands from the registry.
     */
    public void clear() {
        mCommandToActionMap.clear();
        mHelpString.delete(0, mHelpString.length());
    }

    /**
     * Adds a string to the 'shown help' list.
     *
     * @param helpString The text to add.
     */
    public void addToHelp(String helpString) {
        if (mHelpString.length() > 0) {
            mHelpString.append(',');
        }
        mHelpString.append(helpString);
    }

    /**
     * Run a given command if it is registered.
     *
     * @param command The text of the command.
     * @return True if an action was executed, false if not.
     */
    public boolean runCommand(String command) {
        final Runnable action = mCommandToActionMap.get(command);
        if (action == null) {
            return false;
        } else {
            action.run();
            return true;
        }
    }

    /**
     * Get the wear HF directives for adding the registered commands and their associated help
     * commands.
     *
     * @return The directive string.
     */
    public String getCommandsDirective() {
        final ArrayList<String> commandList = new ArrayList<>(mCommandToActionMap.keySet());
        commandList.sort(String::compareToIgnoreCase);
        return "hf_show_help_commands:" + mHelpString + "|" +
                "hf_add_commands: " + String.join("|", mCommandToActionMap.keySet());
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_RESUME)
    public void onResume() {
        mContext.registerReceiver(mCommandReceiver, new IntentFilter(ACTION_SPEECH_EVENT));
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_PAUSE)
    public void onPause() {
        mContext.unregisterReceiver(mCommandReceiver);
    }

    /**
     * When registered, attempts to execute received speech commands.
     */
    private final BroadcastReceiver mCommandReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(final Context context, final Intent intent) {
            if (Objects.equals(intent.getAction(), ACTION_SPEECH_EVENT)) {
                final String command = intent.getStringExtra(EXTRA_COMMAND);
                if (command != null) {
                    runCommand(command);
                }
            }
        }
    };
}
