/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.memory;

import static com.realwear.internal.utils.Utils.concatenateNonEmptyStrings;

import android.content.ContentResolver;
import android.graphics.Bitmap;
import android.graphics.Color;
import android.graphics.pdf.PdfRenderer;
import android.net.Uri;
import android.os.ParcelFileDescriptor;

import androidx.annotation.NonNull;

import com.squareup.picasso.Picasso;
import com.squareup.picasso.Request;
import com.squareup.picasso.RequestHandler;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

public class PDFPageRequestHandler extends RequestHandler {
    public static String SCHEME_PDF_PAGE = "pdf_page";
    @NonNull
    private final ContentResolver mResolver;

    public PDFPageRequestHandler(@NonNull ContentResolver resolver) {
        mResolver = resolver;
    }

    @Override
    public boolean canHandleRequest(Request data) {
        String scheme = data.uri.getScheme();
        return (SCHEME_PDF_PAGE.equals(scheme));
    }

    @Override
    public Result load(Request data, int arg1) {
        final Uri uri = data.uri;
        if (uri.toString().contains("content://")) {
            return handleContentProviderUri(uri);
        } else {
            final String[] info = data.uri.getPath().split(":");
            final int pageNum = Integer.parseInt(info[1]);
            final File file = new File(info[0]);
            Bitmap bitmap = null;

            if (file.exists()) {
                try {
                    bitmap = renderPage(ParcelFileDescriptor.open(file, ParcelFileDescriptor.MODE_READ_ONLY), pageNum);
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }

            if (bitmap == null) {
                return null;
            }

            return new Result(bitmap, Picasso.LoadedFrom.DISK);
        }
    }

    private Result handleContentProviderUri(Uri uri) {
        final String rawUri = uri.toString();
        final String[] uriSections = rawUri.split(":");
        final String uriText = concatenateNonEmptyStrings(":", uriSections[1], uriSections[2]);
        final int pageNumber = Integer.parseInt(uriSections[3]);
        Bitmap bitmap = null;

        try {
            bitmap = renderPage(mResolver.openFileDescriptor(Uri.parse(uriText), "r"), pageNumber);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        if (bitmap == null) {
            return null;
        }

        return new Result(bitmap, Picasso.LoadedFrom.DISK);
    }

    private Bitmap renderPage(ParcelFileDescriptor inputDescriptor, final int pageNum) throws IOException {
        try {
            final PdfRenderer renderer = new PdfRenderer(inputDescriptor);
            renderer.getPageCount();

            try (PdfRenderer.Page page = renderer.openPage(pageNum)) {
                int maxSize = 2048;
                float ratio = Math.min(
                        (float) maxSize / page.getWidth(),
                        (float) maxSize / page.getHeight());
                int width = Math.round(ratio * page.getWidth());
                int height = Math.round(ratio * page.getHeight());

                Bitmap bitmap = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888);

                bitmap.eraseColor(Color.WHITE);
                page.render(bitmap, null, null, PdfRenderer.Page.RENDER_MODE_FOR_DISPLAY);
                return bitmap;
            }
        } catch (Exception e) {
            inputDescriptor.close();
            throw e;
        }
    }
}