/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.memory;

import android.annotation.TargetApi;
import android.graphics.RectF;
import android.os.Build;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.os.Message;
import android.util.Log;

import androidx.annotation.NonNull;

import com.realwear.internal.utils.FlagView;
import com.realwear.internal.utils.FlagWriteView;
import com.realwear.internal.utils.ObservableReadValue;
import com.realwear.internal.utils.ObservableWriteView;
import com.tom_roush.pdfbox.pdmodel.PDDocument;
import com.tom_roush.pdfbox.pdmodel.PDPage;
import com.tom_roush.pdfbox.pdmodel.common.PDRectangle;
import com.tom_roush.pdfbox.text.PDFTextStreamEngine;
import com.tom_roush.pdfbox.text.TextPosition;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

/**
 * Handles the operation of searching a PDF document for given terms.
 */
@TargetApi(value = Build.VERSION_CODES.N)
public class PDFSearchHandler {
    private static final int MSG_PREPARE_SEARCH = 1;
    private static final int MSG_SEARCH_PAGE = 2;
    private static final int MSG_STOP_SEARCH = 3;
    private static final String EXTRA_SEARCH_TERM = "st";
    private static final String EXTRA_SEARCH_PAGE = "sp";
    private static final String TAG = PDFSearchHandler.class.getSimpleName();

    private final Handler mThreadHandler;
    private final Handler mMainHandler = new Handler(Looper.getMainLooper());
    private final AtomicReference<InputStream> mInput = new AtomicReference<>();
    private final AtomicReference<Consumer<SearchResult>> mResultsHandler = new AtomicReference<>();
    private final AtomicReference<PDDocument> mDocument = new AtomicReference<>();

    private final ObservableWriteView<Boolean> mPreparedFlag = new ObservableWriteView<>(false);
    public final ObservableReadValue<Boolean> preparedFlag = mPreparedFlag.getObservableReadValue();
    private final FlagWriteView mErrorFlag = new FlagWriteView();
    public final FlagView errorFlag = mErrorFlag.getView();
    private final FlagWriteView mFinishFlag = new FlagWriteView();
    public final FlagView finishFlag = mFinishFlag.getView();

    public static class SearchResult {
        SearchResult(final int pageNumber, final List<RectF> highlightAreas) {
            this.pageNumber = pageNumber;
            this.highlightAreas = Collections.unmodifiableList(new ArrayList<>(highlightAreas));
        }

        public final int pageNumber;
        public final List<RectF> highlightAreas;

        @NonNull
        @Override
        public String toString() {
            return "SearchResult{" +
                    "pageNumber=" + pageNumber +
                    ", highlightAreas=" + highlightAreas +
                    '}';
        }
    }

    /**
     * Constructor.
     */
    public PDFSearchHandler() {
        final HandlerThread thread = new HandlerThread("PDF Search thread");
        thread.start();
        thread.setPriority(Thread.MIN_PRIORITY);
        mThreadHandler = new Handler(thread.getLooper(), this::handleThreadMessages);
    }

    /**
     * Load a document into memory in preparation for searching.
     *
     * @param documentStream The input stream with the file data.
     */
    public void prepareSearch(InputStream documentStream) {
        mPreparedFlag.update(false);
        mInput.set(documentStream);
        mThreadHandler.sendEmptyMessage(MSG_PREPARE_SEARCH);
    }

    /**
     * Start a search process for a given search term.
     *
     * @param textToFind    The text to be found.
     * @param resultHandler The handler to receive the results. Note: Calls to this handler are
     *                      processed in the UX thread (i.e. Looper.getMainLooper()).
     */
    public void beginSearch(String textToFind, Consumer<SearchResult> resultHandler) {
        final Message message = mThreadHandler.obtainMessage(MSG_SEARCH_PAGE);
        message.getData().putString(EXTRA_SEARCH_TERM, textToFind);
        message.getData().putInt(EXTRA_SEARCH_PAGE, 0);
        mResultsHandler.set(resultHandler);
        mThreadHandler.sendMessage(message);
    }

    /**
     * Sends a message to cancel any ongoing search. Will not happen instantly.
     */
    public void cancelSearch() {
        mThreadHandler.sendEmptyMessage(MSG_STOP_SEARCH);
    }

    /**
     * Dispatches all messages on the search thread.
     *
     * @param messageIn The message received.
     * @return True - specified by the Handler interface.
     */
    private boolean handleThreadMessages(Message messageIn) {
        switch (messageIn.what) {
            case MSG_PREPARE_SEARCH:
                doPrepareSearch();
                break;
            case MSG_SEARCH_PAGE:
                final String searchTerm = messageIn.getData().getString(EXTRA_SEARCH_TERM);
                final int pageNumber = messageIn.getData().getInt(EXTRA_SEARCH_PAGE);
                if (doSearchPage(searchTerm, pageNumber)) {
                    final Message newMessage = mThreadHandler.obtainMessage(MSG_SEARCH_PAGE);
                    newMessage.getData().putString(EXTRA_SEARCH_TERM, searchTerm);
                    newMessage.getData().putInt(EXTRA_SEARCH_PAGE, pageNumber + 1);
                    mThreadHandler.sendMessageDelayed(newMessage, 500);
                } else {
                    mFinishFlag.trigger();
                }
                break;
            case MSG_STOP_SEARCH:
                mThreadHandler.removeMessages(MSG_PREPARE_SEARCH);
                mThreadHandler.removeMessages(MSG_SEARCH_PAGE);
                mFinishFlag.trigger();
                break;
        }
        return true;
    }

    /**
     * Actually does the prepare-search task - to be called from the worker thread.
     */
    private void doPrepareSearch() {
        final InputStream input = mInput.getAndSet(null);
        if (input == null) {
            return;
        }
        final PDDocument document = mDocument.getAndSet(null);
        if (document != null) {
            try {
                document.close();
            } catch (Exception e) {
                Log.w(TAG, "Exception closing document", e);
            }
        }
        try {
            mDocument.set(PDDocument.load(input));
            mMainHandler.post(() -> mPreparedFlag.update(true));
        } catch (IOException e) {
            Log.w(TAG, "Failed to load document for searching", e);
            mMainHandler.post(mErrorFlag::trigger);
        } finally {
            try {
                input.close();
            } catch (IOException e) {
                Log.d(TAG, "Exception closing input", e);
            }
        }
    }

    /**
     * Searches a page for the provided text to be found. Calls the handler for processing
     * results on a match being found.
     *
     * @param textToFind The text to be found.
     * @param pageNumber The page number to be searched.
     * @return True if the search should be continued after this point.
     */
    private boolean doSearchPage(String textToFind, final int pageNumber) {
        final Consumer<SearchResult> resultsHandler = mResultsHandler.get();
        final PDDocument currentDocument = mDocument.get();
        if (resultsHandler == null) {
            Log.w(TAG, "Search not started as no handler for results found");
            return false;
        }
        try {
            CustomTextScraper searcher = new CustomTextScraper(textToFind,
                    rectangles -> {
                        final SearchResult result = new SearchResult(pageNumber,
                                rectangles);
                        mMainHandler.post(() -> resultsHandler.accept(result));
                    });

            if (pageNumber < currentDocument.getNumberOfPages()) {
                final PDPage page = currentDocument.getPage(pageNumber);
                searcher.processPage(page);
                return true;
            } else {
                return false;
            }
        } catch (IOException e) {
            Log.w(TAG, "Error in searching", e);
            mErrorFlag.trigger();
            return false;
        }
    }

    private static class CustomTextScraper extends PDFTextStreamEngine {
        @NonNull
        private final Consumer<List<RectF>> mResultHandler;
        @NonNull
        private final String mSearchTerm;
        private StringBuilder mResultText = new StringBuilder();
        private final List<RectF> mResultArea = new LinkedList<>();
        private PDRectangle mCurrentPageSize;

        /**
         * Constructor.
         */
        CustomTextScraper(@NonNull String searchTerm,
                          @NonNull Consumer<List<RectF>> resultHandler) throws IOException {
            super();
            mSearchTerm = searchTerm.toLowerCase();
            mResultHandler = resultHandler;
        }

        @Override
        public void processPage(final PDPage page) throws IOException {
            mCurrentPageSize = page.getBBox();
            super.processPage(page);
        }

        @Override
        protected void processTextPosition(final TextPosition text) {
            mResultText.append(text.getUnicode().toLowerCase());
            final String currentResult = mResultText.toString();
            mResultArea.add(new RectF((text.getX() - mCurrentPageSize.getLowerLeftX()) / mCurrentPageSize.getWidth(),
                    ((text.getY() + mCurrentPageSize.getLowerLeftY()) - text.getHeight() * 1.3f) / mCurrentPageSize.getHeight(),
                    ((text.getX() - mCurrentPageSize.getLowerLeftX()) + text.getWidth()) / mCurrentPageSize.getWidth(),
                    ((text.getY() + mCurrentPageSize.getLowerLeftY()) + text.getHeight() * 0.4f) / mCurrentPageSize.getHeight()));
            if (mSearchTerm.equals(currentResult)) {
                mResultHandler.accept(mResultArea);
                mResultArea.clear();
                mResultText = new StringBuilder();
            } else {
                while (mResultText.length() > 0 && !mSearchTerm.startsWith(mResultText.toString())) {
                    mResultText.delete(0, 1);
                    mResultArea.remove(0);
                }
            }
            super.processTextPosition(text);
        }
    }
}
