/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.memory;

import android.app.Activity;
import android.content.Context;
import android.content.ContextWrapper;
import android.content.SharedPreferences;

import java.util.ArrayList;
import java.util.List;

/**
 * Represents the interface to the stored recent searches.
 */
public class RecentSearches {
    private static final int MAX_SEARCH_HISTORY_SIZE = 10;
    private static final String PREFERENCE_NAME = "RecentSearchTerm%d";

    private List<String> mCachedSearches;
    private final Activity mContext;

    /**
     * Constructor.
     *
     * @param context Operating context.
     */
    public RecentSearches(final Context context) {
        mContext = findActivity(context);
    }

    /**
     * Returns a list of recent searches, starting with the most recent to least recent.
     *
     * @return The list of terms.
     */
    public List<String> getRecentSearches() {
        if (mCachedSearches == null) {
            mCachedSearches = new ArrayList<>(MAX_SEARCH_HISTORY_SIZE);
            final SharedPreferences prefs = mContext.getPreferences(Context.MODE_PRIVATE);
            for (int i = 0; i < MAX_SEARCH_HISTORY_SIZE; i++) {
                final String entry = prefs.getString(String.format(PREFERENCE_NAME, i), null);
                if (entry != null) {
                    mCachedSearches.add(entry);
                }
            }
        }
        return new ArrayList<>(mCachedSearches);
    }

    /**
     * Adds a search term as the newest search term.
     *
     * @param searchTerm The new search term.
     */
    public void addRecentSearch(String searchTerm) {
        mCachedSearches.remove(searchTerm);
        mCachedSearches.add(0, searchTerm);
        while (mCachedSearches.size() > 10) {
            mCachedSearches.remove(10);
        }
        final SharedPreferences prefs = mContext.getPreferences(Context.MODE_PRIVATE);
        for (int i = 0; i < mCachedSearches.size(); i++) {
            final SharedPreferences.Editor editor = prefs.edit();
            editor.putString(String.format(PREFERENCE_NAME, i), mCachedSearches.get(i));
            editor.apply();
        }
    }

    /**
     * Find the parent activity from a given context.
     *
     * @param context The context to start at.
     * @return The activity found.
     */
    private static Activity findActivity(Context context) {
        if (context instanceof Activity) {
            return (Activity) context;
        } else if (context instanceof ContextWrapper) {
            return findActivity(((ContextWrapper) context).getBaseContext());
        } else {
            throw new IllegalStateException("Could not find a base activity for this context");
        }
    }
}
