/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.utils;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.util.Log;

import com.realwear.documentviewer.object.Bookmark;

import java.util.ArrayList;
import java.util.List;

public class DataManager extends SQLiteOpenHelper {
    private static final int DATABASE_VERSION = 2;
    private static final String DATABASE_NAME = "DataManager";
    private static final String TAG = DataManager.class.getName();

    private static final String TABLE_BOOKMARKS = "bookmarks";
    private static final String KEY_ID = "id";
    private static final String KEY_LOCATION = "document_location";
    private static final String KEY_BOOKMARK_INDEX = "bookmark_index";
    private static final String KEY_DOCUMENT_MODIFIED = "bookmark_mod";

    public DataManager(Context context) {
        super(context, DATABASE_NAME, null, DATABASE_VERSION);
        //3rd argument to be passed is CursorFactory instance
    }

    // Creating Tables
    @Override
    public void onCreate(SQLiteDatabase db) {
        String CREATE_CONTACTS_TABLE = "CREATE TABLE " + TABLE_BOOKMARKS + "("
                + KEY_ID + " INTEGER PRIMARY KEY," + KEY_LOCATION + " TEXT,"
                + KEY_BOOKMARK_INDEX + " INTEGER,"
                + KEY_DOCUMENT_MODIFIED + " TEXT" + ")";
        db.execSQL(CREATE_CONTACTS_TABLE);
    }

    // Upgrading database
    @Override
    public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
        // Drop older table if existed
        db.execSQL("DROP TABLE IF EXISTS " + TABLE_BOOKMARKS);

        // Create tables again
        onCreate(db);
    }

    public void addBookmark(Bookmark bookmark) {
        try (SQLiteDatabase db = this.getWritableDatabase()) {
            ContentValues values = new ContentValues();
            values.put(KEY_LOCATION, bookmark.getDocumentLocation());
            values.put(KEY_BOOKMARK_INDEX, bookmark.getBookmarkNumber());
            values.put(KEY_DOCUMENT_MODIFIED, bookmark.getDateMod());
            // Inserting Row
            db.insert(TABLE_BOOKMARKS, null, values);
        }
    }

    public List<Bookmark> getAllBookmarks(String docLocation, String fileMod) {
        final List<Bookmark> contactList = new ArrayList<>();
        // Select All Query
        String selectQuery = "SELECT * FROM " + TABLE_BOOKMARKS + " where document_location = ? and bookmark_mod = ?";

        try (SQLiteDatabase db = this.getWritableDatabase();
             Cursor cursor = db.rawQuery(selectQuery, new String[]{docLocation, fileMod})) {

            // looping through all rows and adding to list
            if (cursor.moveToFirst()) {
                do {
                    Bookmark contact = new Bookmark(cursor.getInt(0), cursor.getInt(2), cursor.getString(1), cursor.getString(3));
                    // Adding contact to list
                    contactList.add(contact);
                } while (cursor.moveToNext());
            }
        } catch (Exception e) {
            Log.d(TAG, "Failed to read the bookmarks from the database: " + docLocation, e);
        }

        // return contact list
        return contactList;
    }

    public void deleteBookmark(Bookmark contact) {
        SQLiteDatabase db = this.getWritableDatabase();
        db.delete(TABLE_BOOKMARKS, KEY_BOOKMARK_INDEX + " = ? and " + KEY_LOCATION + " = ?" + "and " + KEY_DOCUMENT_MODIFIED + " = ?",
                new String[]{String.valueOf(contact.getBookmarkNumber()), String.valueOf(contact.getDocumentLocation()), String.valueOf(contact.getDateMod())});
        db.close();
    }
}
