/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.utils

import android.Manifest
import android.app.Activity
import android.content.pm.PackageManager
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat

/**
 * Utility class for using Android permissions.
 */
internal object PermissionUtils {
    private const val REQUEST_CODE = 123
    private val APP_PERMISSIONS = arrayOf(
        Manifest.permission.WRITE_EXTERNAL_STORAGE
    )

    /**
     * Check that the application's [activity]s has the permissions it needs to function.
     */
    fun checkAppPermissions(activity: Activity): Boolean {
        return APP_PERMISSIONS.all {
            ContextCompat.checkSelfPermission(activity, it) == PackageManager.PERMISSION_GRANTED
        }
    }

    /**
     * Request all the permissions the app [activity]s need to function.
     * Listen to the response using the activities onRequestPermissionsResult method. Validate
     * that the permissions were granted using [validatePermissionResult].
     */
    fun requestAppPermissions(activity: Activity) {
        val permissionsRequired = APP_PERMISSIONS.filter {
            ContextCompat.checkSelfPermission(activity, it) != PackageManager.PERMISSION_GRANTED
        }

        if (permissionsRequired.isNotEmpty()) {
            ActivityCompat.requestPermissions(
                activity, permissionsRequired.toTypedArray(), REQUEST_CODE
            )
        }
    }

    /**
     * Validate that all the requested permissions were granted by the user.
     */
    fun validatePermissionResult(requestCode: Int, grantResults: IntArray): Boolean {
        if (requestCode != REQUEST_CODE || grantResults.isEmpty()) {
            // The result array will be empty if the request is cancelled.
            return false
        }

        return grantResults.all { it == PackageManager.PERMISSION_GRANTED }
    }
}