/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.views;

import static com.realwear.internal.utils.Functional.runIfParameterNotNull;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.PorterDuff;
import android.graphics.RectF;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.util.AttributeSet;
import android.view.TextureView;

import androidx.annotation.Nullable;
import androidx.lifecycle.Lifecycle;
import androidx.lifecycle.LifecycleObserver;
import androidx.lifecycle.OnLifecycleEvent;

import com.realwear.internal.utils.Flag;
import com.realwear.view.styles.Theme;
import com.realwear.view.styles.ThemeAwareObject;

import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

public class BitmapDisplay extends TextureView implements ThemeAwareObject, LifecycleObserver {
    public static final int HIGHLIGHT_COLOUR = Color.argb(100, 255, 255, 100);
    private static final float INTER_ZOOM_FACTOR = 1.7320508076f;

    /**
     * Matrices used for Scaling, Positioning and Centering
     */
    private Matrix mViewMatrix = new Matrix();
    private Matrix mViewPosition = new Matrix();
    private Matrix mViewScale = new Matrix();
    private Matrix mViewPostTranslate = new Matrix();
    private Matrix mFinderPostTranslate = new Matrix();
    /**
     * View Postion
     */
    private float mViewPositionX = 0;
    private float mViewPositionY = 0;
    private float mScalePercent = 0;
    private int mZoomLevel = 0;
    private Bitmap mBitmap;
    private HandlerThread mRenderThread;
    private boolean mLocked;
    private final Paint mViewPaint;
    private final Paint mHighlightPaint = new Paint();

    @Nullable
    private DisplayPreview displayPreview;

    private int mBackgroundColour = Theme.DEFAULT.getColor(Theme.ColorType.SURFACE);
    private float mZoomFactor = 1;
    private final List<RectF> mHighlightRectangles =
            new CopyOnWriteArrayList<>();
    private final Flag mRerenderFlag = new Flag();
    private Handler mRenderHandler;

    /**
     * Clear all highlight rectangles from the bitmap.
     */
    public void clearHighlightRectangles() {
        mHighlightRectangles.clear();
        doRender();
    }

    /**
     * Adds a collection of highlight rectangles to the bitmap.
     *
     * @param rectangles The rectangles to add.
     */
    public void addHighlightRectangles(final List<RectF> rectangles) {
        mHighlightRectangles.addAll(rectangles);
        doRender();
    }

    public BitmapDisplay(Context context) {
        this(context, null);
    }

    public BitmapDisplay(Context context, AttributeSet attrs) {
        this(context, attrs, 0);
    }

    public BitmapDisplay(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);

        mViewPaint = new Paint(Paint.FILTER_BITMAP_FLAG);
        mViewPaint.setAntiAlias(true);
        mViewPaint.setFilterBitmap(true);
        mViewPaint.setDither(true);

        mHighlightPaint.setStyle(Paint.Style.FILL);
        mHighlightPaint.setColor(HIGHLIGHT_COLOUR);

        mRerenderFlag.addListener(this::doRender);

        onResume();
    }

    public void setBitmap(Bitmap bitmap) {
        mBitmap = bitmap;

        resetView(true, true);
    }

    public void setLocation(int x, int y) {
        if (x > mBitmap.getWidth()) {
            x = mBitmap.getWidth();
        }
        if (y > mBitmap.getHeight()) {
            y = mBitmap.getHeight();
        }

        mViewPositionX = x;
        mViewPositionY = y;

        if (displayPreview != null) {
            displayPreview.setPreviewPostion(mViewPositionX, mViewPositionY);
        }

        calculateMatrix();
    }

    public void translateView(float x, float y) {
        if (!mLocked && mBitmap != null && mZoomLevel > 0) {
            float speed;
            switch (mZoomLevel) {
                case 1:
                    speed = 2.5f;
                    break;
                case 2:
                    speed = 3.0f;
                    break;
                case 3:
                case 4:
                    speed = 3.5f;
                    break;
                default:
                    speed = 0;
                    break;
            }

            float tempmHX = mViewPositionX + (x * speed);
            float tempmHY = mViewPositionY + (y * speed);

            if (tempmHX < mBitmap.getWidth() && tempmHX > 0) {
                mViewPositionX += x * speed;
            }

            if (tempmHY < mBitmap.getHeight() && tempmHY > 0) {
                mViewPositionY += y * speed;
            }

            runIfParameterNotNull(previewer -> previewer.setPreviewPostion(mViewPositionX, mViewPositionY), displayPreview);

            calculateMatrix();
        }
    }

    public void setZoomLevel(int level) {
        mZoomLevel = level;

        float scaleRatio = calculateScaleRatio();
        if (mZoomLevel != 0) {

            mViewScale.setScale(scaleRatio, scaleRatio);
            calculateMatrix();
        } else {
            resetView(true, true);
        }

        if (displayPreview != null) {
            displayPreview.setZoomAmount(mZoomFactor);
        }
    }

    private float calculateScaleRatio() {
        mZoomFactor = (float) Math.pow(INTER_ZOOM_FACTOR, mZoomLevel);
        return mScalePercent * mZoomFactor;
    }

    public int getZoomLevel() {
        return mZoomLevel;
    }

    public void lockView(boolean lock) {
        mLocked = lock;
    }

    public boolean isViewLocked() {
        return mLocked;
    }

    /**
     * Resets the View to center point
     */
    public void resetView(boolean recalculate, boolean reposition) {
        if (mBitmap == null) {
            return;
        }

        if (reposition) {
            mViewPositionX = mBitmap.getWidth() / 2;
            mViewPositionY = mBitmap.getHeight() / 2;
        }

        // "fit" to the screen
        if (recalculate) {
            float zoomX = (float) getWidth() / ((float) mBitmap.getWidth());
            float zoomY = (float) getHeight() / ((float) mBitmap.getHeight());
            mScalePercent = Math.min(zoomX, zoomY);

            float scaleRatio = calculateScaleRatio();
            // set View to default zoom level
            mViewScale.setScale(scaleRatio, scaleRatio);

            // Set View viewers center point
            mViewPostTranslate.setTranslate(getWidth() / 2, getHeight() / 2);

            mFinderPostTranslate.setTranslate(getWidth() / 2, getHeight() / 2);

            if (displayPreview != null) {
                displayPreview.setZoomAmount(mZoomFactor);
                displayPreview.setPreviewPostion(mViewPositionX, mViewPositionY);
            }
        }

        calculateMatrix();
    }

    private void calculateMatrix() {
        // Set position
        mViewPosition.setTranslate(-mViewPositionX, -mViewPositionY);
        mViewMatrix.setConcat(mViewScale, mViewPosition);
        mViewMatrix.setConcat(mViewPostTranslate, mViewMatrix);

        doRender();
    }

    public void setDisplayPreview(@Nullable DisplayPreview displayPreview) {
        this.displayPreview = displayPreview;
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_PAUSE)
    public void onPause() {
        if (mRenderThread != null) {
            mRenderHandler = null;
            mRenderThread.quitSafely();
            mRenderThread = null;
        }
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_RESUME)
    public void onResume() {
        if (mRenderThread == null) {
            mRenderThread = new HandlerThread("Renderer thread");
            mRenderThread.start();
            mRenderHandler = new Handler(mRenderThread.getLooper());
        }

        doRender();
    }

    @Override
    public void handleThemeUpdate(Theme theme) {
        mBackgroundColour = theme.getColor(Theme.ColorType.SURFACE);
    }

    /**
     * Handles Canvas rendering and SurfaceTexture callbacks.
     * <p>
     * We don't create a Looper, so the SurfaceTexture-by-way-of-TextureView callbacks
     * happen on the UI thread.
     */
    private void doRender() {
        if (mRenderThread == null) {
            return;
        }
        if (Looper.myLooper() != mRenderThread.getLooper()) {
            mRenderHandler.post(this::doRender);
            return;
        }

        Canvas canvas = null;
        try {
            canvas = this.lockCanvas();

            if (canvas == null) {
                mRenderHandler.postDelayed(this::doRender, 10);
            } else {
                // Draw the scene
                canvas.drawColor(0, PorterDuff.Mode.CLEAR);
                canvas.drawColor(mBackgroundColour);
                canvas.setMatrix(mViewMatrix);

                if (mBitmap != null) {
                    canvas.drawBitmap(mBitmap, 0, 0, mViewPaint);
                    for (RectF rectangle : mHighlightRectangles) {
                        canvas.drawRect(rectangle.left * mBitmap.getWidth(),
                                rectangle.top * mBitmap.getHeight(),
                                rectangle.right * mBitmap.getWidth(),
                                rectangle.bottom * mBitmap.getHeight(),
                                mHighlightPaint);
                    }
                }
            }
        } finally {
            if (canvas != null) {
                this.unlockCanvasAndPost(canvas);
            }
        }
    }
}
