/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.views;

import android.content.Context;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.util.AttributeSet;
import android.view.LayoutInflater;
import android.view.View;
import android.view.WindowManager;
import android.widget.FrameLayout;
import android.widget.TextView;

import androidx.recyclerview.widget.GridLayoutManager;
import androidx.recyclerview.widget.RecyclerView;

import com.realwear.documentviewer.R;
import com.realwear.documentviewer.adapters.BookmarkAdapter;
import com.realwear.documentviewer.object.Bookmark;
import com.realwear.headtrackerhelperlib.HeadTrackerHelper;
import com.realwear.view.ContainedButton;

import java.util.List;

public class BookmarkOverlay extends FrameLayout implements SensorEventListener {
    private RecyclerView mDocumentPreviewPages;
    private BookmarkAdapter mAdapter;
    private TextView mStatusText;
    private ContainedButton mPageRange;

    private HeadTrackerHelper htHelper;

    private SensorManager mSensorManager;
    private Sensor mRotationSensor;

    public BookmarkOverlay(Context context) {
        super(context);
        init();
    }

    public BookmarkOverlay(Context context, AttributeSet attrs) {
        super(context, attrs);
        init();
    }

    public BookmarkOverlay(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        init();
    }

    private void init() {
        LayoutInflater.from(getContext()).inflate(R.layout.bookmark_overlay, this);

        mPageRange = findViewById(R.id.page_range);

        mDocumentPreviewPages = findViewById(R.id.documentPreview);
        mDocumentPreviewPages.setHasFixedSize(false);

        GridLayoutManager layoutManager = new GridLayoutManager(getContext(), 2);

        layoutManager.setOrientation(GridLayoutManager.HORIZONTAL);
        mAdapter = new BookmarkAdapter();
        mDocumentPreviewPages.setLayoutManager(layoutManager);
        mDocumentPreviewPages.setAdapter(mAdapter);
        mStatusText = findViewById(R.id.status_text);

        htHelper = new HeadTrackerHelper(
                (WindowManager) getContext().getSystemService(Context.WINDOW_SERVICE)
        );
        mSensorManager = (SensorManager) getContext().getSystemService(Context.SENSOR_SERVICE);
        mRotationSensor = mSensorManager.getDefaultSensor(Sensor.TYPE_ROTATION_VECTOR);
    }

    public void onResume() {
        if (isVisible()) {
            mSensorManager.registerListener(this, mRotationSensor, SensorManager.SENSOR_DELAY_GAME);
        }
    }

    public void onPause() {
        if (isVisible()) {
            mSensorManager.unregisterListener(this);
        }
    }

    public void onRefresh(List<Bookmark> bookmarks, int pageCount) {
        if (bookmarks.size() > 0) {
            mStatusText.setVisibility(View.GONE);
        } else {
            mStatusText.setVisibility(View.VISIBLE);
        }

        mPageRange.removeIcon();
        mPageRange.setText(getContext().getString(R.string.select_page_range, 1, pageCount));
        mPageRange.setClickableVoiceCommand("");

        mAdapter.setBookmarks(bookmarks);

        // divide the number of bookmarks by 2, as there are 2 rows, and then divide by 2 again to
        // get the middle element
        mDocumentPreviewPages.scrollToPosition(mAdapter.getItemCount() / 4);
    }

    public boolean isVisible() {
        return getVisibility() == View.VISIBLE;
    }

    public void show(List<Bookmark> bookmarks, int pageCount) {
        setVisibility(View.VISIBLE);
        onRefresh(bookmarks, pageCount);
        onResume();
    }

    public void hide() {
        onPause();
        setVisibility(View.INVISIBLE);
    }

    @Override
    public void onSensorChanged(SensorEvent event) {
        if (event.sensor == mRotationSensor) {
            final HeadTrackerHelper.Rotation rotation =
                    htHelper.calculateXY(event.values.clone());
            if (mDocumentPreviewPages != null) {
                mDocumentPreviewPages.scrollBy((int) rotation.getX(), 0);
            }
        }
    }

    @Override
    public void onAccuracyChanged(Sensor sensor, int accuracy) {
        // No-op
    }
}
