/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer.views;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Rect;
import android.graphics.RectF;
import android.os.Build;
import android.util.AttributeSet;
import android.util.DisplayMetrics;
import android.view.View;

import com.realwear.documentviewer.R;

public class DisplayPreview extends View {
    /*
     * How much have we scaled the document in order to fit into mini map
     */
    private float mScalePercent = 0;

    /*
     * Position of mini map tracker
     */
    private float mPreviewPosX = 0;
    private float mPreviewPosY = 0;

    /*
     * Matrix used for scaling image
     */
    private Matrix mPreviewMatrix = new Matrix();

    /*
     * Mini map document preview image
     */
    private Bitmap mPreviewBitmap = null;

    /*
     * Mini map tracker paint brush
     */
    private Paint mPreviewStroke = null;

    /*
     * Mini map tracker size and location
     */
    private RectF mPreviewRect = null;
    private Paint mOuterStroke;
    private Matrix mTranslateMatrix = new Matrix();
    private Matrix mScaleMatrix = new Matrix();
    private Matrix mViewPostTranslate = new Matrix();

    private Rect mOuterRect;

    private float mZoomAmount = 1.0f;

    public DisplayPreview(Context context) {
        super(context);
        init();
    }

    public DisplayPreview(Context context, AttributeSet attrs) {
        super(context, attrs);
        init();
    }

    public DisplayPreview(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        init();
    }

    public DisplayPreview(Context context, AttributeSet attrs, int defStyleAttr, int defStyleRes) {
        super(context, attrs, defStyleAttr, defStyleRes);
        init();
    }

    private void init() {
        // Create our paint brush
        float screenDpiRatio = ((float) getContext().getResources().getDisplayMetrics().densityDpi /
                DisplayMetrics.DENSITY_DEFAULT);

        mPreviewStroke = new Paint();
        mPreviewStroke.setStyle(Paint.Style.STROKE);
        mPreviewStroke.setStrokeWidth(4 * screenDpiRatio);
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            mPreviewStroke.setColor(getContext().getColor(R.color.radio));
        }

        mOuterStroke = new Paint();
        mOuterStroke.setStyle(Paint.Style.STROKE);
        mOuterStroke.setStrokeWidth(5 * screenDpiRatio);
        mOuterStroke.setColor(Color.BLACK);

        mOuterRect = new Rect(2, 2, getWidth() - 2, getHeight() - 2);
    }

    @Override
    protected void onSizeChanged(int w, int h, int oldw, int oldh) {
        super.onSizeChanged(w, h, oldw, oldh);

        mOuterRect = new Rect(2, 2, w - 2, h - 2);

        /*
         * Scale the preview image by the largest size
         */
        if (mPreviewBitmap != null) {
            float scaleXPercent = (float) w / (float) mPreviewBitmap.getWidth();
            float scaleYPercent = (float) h / (float) mPreviewBitmap.getHeight();
            mScalePercent = (Math.min(scaleXPercent, scaleYPercent));

            float viewPositionX = mPreviewBitmap.getWidth() / 2;
            float viewPositionY = mPreviewBitmap.getHeight() / 2;
            mViewPostTranslate.setTranslate(w / 2, h / 2);
            mTranslateMatrix.setTranslate(-viewPositionX, -viewPositionY);
            mScaleMatrix.setScale(mScalePercent, mScalePercent);
            mPreviewMatrix.setConcat(mScaleMatrix, mTranslateMatrix);
            mPreviewMatrix.setConcat(mViewPostTranslate, mPreviewMatrix);
        }

        // Set the position of our tracker
        setPreviewPostion(0, 0);
    }

    public void setPreviewBitmap(Bitmap bitmap) {
        mPreviewBitmap = bitmap;

        onSizeChanged(this.getWidth(), this.getHeight(), 0, 0);

        updateRect();
    }

    public void setPreviewPostion(float x, float y) {
        mPreviewPosX = (x * mScalePercent);
        mPreviewPosY = (y * mScalePercent);

        updateRect();
    }

    public void updateRect() {
        float btmScaleWidth;
        float btmScaleHeight;
        if (mPreviewBitmap != null) {
            // get the image width within the preview
            btmScaleWidth = mPreviewBitmap.getWidth() * mScalePercent;
            btmScaleHeight = mPreviewBitmap.getHeight() * mScalePercent;
        } else {
            btmScaleWidth = getWidth();
            btmScaleHeight = getHeight();
        }
        // calculate the start positions within the preview
        float startX = getWidth() / 2.f - (btmScaleWidth / 2.f) + mPreviewPosX;
        float startY = getHeight() / 2.f - (btmScaleHeight / 2.f) + mPreviewPosY;
        // calculate the size of the preview
        float prevWidth = getWidth() / mZoomAmount;
        float prevHeight = getHeight() / mZoomAmount;

        // calculate the position of the preview
        float left = startX - (prevWidth / 2.f);
        float top = startY - (prevHeight / 2.f);
        float right = left + prevWidth;
        float bottom = top + prevHeight;
        mPreviewRect = new RectF(left, top, right, bottom);

        invalidate();
    }

    public void setZoomAmount(float amount) {
        mZoomAmount = amount;
        updateRect();
    }

    public void onDraw(Canvas canvas) {
        super.onDraw(canvas);

        if (mPreviewBitmap != null) {
            canvas.drawBitmap(mPreviewBitmap, mPreviewMatrix, null);
        }

        canvas.drawRect(mOuterRect, mOuterStroke);
        canvas.drawRect(mPreviewRect, mPreviewStroke);
    }
}