/*
 * RealWear Development Software, Source Code and Object Code.
 * (c) RealWear, Inc. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Contact info@realwear.com for further information about the use of this code.
 */

package com.realwear.documentviewer;

import static androidx.test.core.app.ApplicationProvider.getApplicationContext;
import static com.google.common.truth.Truth.assertThat;
import static com.realwear.documentviewer.MainActivity.SHOW_PREVIOUS_NEXT;

import android.content.Intent;
import android.net.Uri;
import android.view.View;
import android.view.ViewGroup;
import android.widget.TextView;

import androidx.annotation.StringRes;
import androidx.lifecycle.Lifecycle;
import androidx.test.core.app.ActivityScenario;
import androidx.test.rule.GrantPermissionRule;

import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.robolectric.RobolectricTestRunner;
import org.robolectric.annotation.Config;
import org.robolectric.annotation.LooperMode;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Predicate;

/**
 * To work on unit tests, switch the Test Artifact in the Build Variants view.
 */
@LooperMode(LooperMode.Mode.PAUSED)
@Config(application = DocumentApplicationTest.class)
@RunWith(RobolectricTestRunner.class)
public class BookmarkTests {
    private static final String JSON_PAGE = "page";
    private static final String JSON_ZOOM = "zoom";
    private static final String JSON_X_POSITION = "x";
    private static final String JSON_Y_POSITION = "y";
    private static final String IS_BOOKMARK = "is_bookmark";
    private static final String EXTRA_FILE_PATH = "file_path";

    @Rule
    public GrantPermissionRule mRuntimePermissionRule = GrantPermissionRule.grant(android.Manifest.permission.WRITE_EXTERNAL_STORAGE);

    @Test
    public void testOpenBookmark() throws IOException {
        Intent intent = new Intent(Intent.ACTION_VIEW);
        intent.addCategory(Intent.CATEGORY_DEFAULT);
        intent.setDataAndType(Uri.parse("./test.pdf"), "application/pdf");
        assertThat(intent.getData()).isNotNull();
        intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        intent.putExtra(EXTRA_FILE_PATH, "./test.pdf");
        intent.putExtra(SHOW_PREVIOUS_NEXT, false);
        intent.putExtra(JSON_PAGE, "1");
        intent.putExtra(JSON_ZOOM, "2");
        intent.putExtra(JSON_X_POSITION, "0");
        intent.putExtra(JSON_Y_POSITION, "0");
        intent.putExtra(IS_BOOKMARK, true);

        try (ActivityScenario<MainActivity> scenario = ActivityScenario.launch(intent)) {
            scenario.moveToState(Lifecycle.State.CREATED);
            scenario.moveToState(Lifecycle.State.STARTED);
            scenario.onActivity(activity -> activity.mDocumentTarget.onBitmapLoaded(null, null));
            scenario.moveToState(Lifecycle.State.RESUMED);
            scenario.onActivity(activity -> {
                assertThat(countControlWithText(activity, R.string.freeze_document)).isEqualTo(0);
                assertThat(countControlWithText(activity, R.string.control_document)).isEqualTo(1);
            });
            scenario.moveToState(Lifecycle.State.STARTED);
            scenario.moveToState(Lifecycle.State.DESTROYED);
        }
    }

    private int countControlWithText(MainActivity activity, @StringRes final int stringId) {
        final String TARGET_TEXT = getApplicationContext().getString(stringId);
        List<TextView> result = findAllMatchingViews(activity, tv -> TARGET_TEXT.equalsIgnoreCase(tv.getText().toString()), TextView.class);
        return result.size();
    }

    private <T extends View> List<T> findAllMatchingViews(MainActivity activity, Predicate<T> filterTest, Class<? extends T> targetType) {
        ViewGroup parent = activity.findViewById(android.R.id.content);
        return findAllMatchingViews(parent, filterTest, targetType);
    }

    private <T extends View> List<T> findAllMatchingViews(ViewGroup startPoint, Predicate<T> filterTest, Class<? extends T> targetType) {
        List<T> results = new LinkedList<>();
        for (int i = 0; i < startPoint.getChildCount(); i++) {
            View child = startPoint.getChildAt(i);
            if (child instanceof ViewGroup) {
                results.addAll(findAllMatchingViews((ViewGroup) child, filterTest, targetType));
            } else if (targetType.isInstance(child) && filterTest.test((T) child)) {
                results.add((T) child);
            }
        }
        return results;
    }
}